"""
=========================================================================
net_sinks
=========================================================================
Test sinks for networks, which does not check the order of the received
messages.

Author : Yanghui Ou
  Date : Apr 30, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, RecvRTL2SendCL, enrdy_to_str

#-------------------------------------------------------------------------
# TestNetSinkCL
#-------------------------------------------------------------------------

class TestNetSinkCL( Component ):

  def construct( s,
    Type,
    msgs,
    initial_delay=0,
    interval_delay=0,
    arrival_time=None,
    match_func=lambda a, b : a.src == b.src and a.dst==b.dst and a.payload == b.payload,
  ):

    s.recv.Type = Type

    # [msgs] and [arrival_time] must have the same length.
    if arrival_time is not None:
      assert len( msgs ) == len( arrival_time )

    s.idx          = 0
    s.cycle_count  = 0
    s.msgs         = list( msgs )
    s.nmsgs        = len( msgs )
    s.arrival_time = None if arrival_time is None else \
                     list( arrival_time )
    s.perf_regr    = True if arrival_time is not None else False
    s.error_msg    = ""
    # TODO: maybe make this a parameter
    s.match_func   = match_func

    s.initial_count  = initial_delay
    s.interval_delay = interval_delay
    s.interval_count = 0

    s.recv_msg    = None
    s.recv_called = False
    s.recv_rdy    = False
    s.trace_len   = len( str( s.msgs[0] ) ) if len(s.msgs) != 0 else 0
#    s.trace_len   = len( str( s.msgs[0] ) )

    @s.update
    def up_sink_count():
      if s.error_msg:
        raise Exception( s.error_msg )
      if not s.reset:
        s.cycle_count += 1
      else:
        s.cycle_count = 0

      # if recv was called in previous cycle
      if s.recv_called:
        s.interval_count = s.interval_delay

      elif s.initial_count != 0:
        s.initial_count -= 1

      elif s.interval_count != 0:
        s.interval_count -= 1

      else:
        s.interval_count = 0

      s.recv_called = False
      s.recv_msg    = None
      s.recv_rdy    = s.initial_count == 0 and s.interval_count == 0

    s.add_constraints(
      U( up_sink_count ) < M( s.recv ),
      U( up_sink_count ) < M( s.recv.rdy )
    )

  @non_blocking( lambda s: s.initial_count==0 and s.interval_count==0 )
  def recv( s, msg ):

    s.recv_msg = msg

    # Sanity check
    if s.idx >= s.nmsgs:
      s.error_msg = ("""
Test Sink received more msgs than expected
Received : {}
""".format( msg ) )

    # Check correctness first
    if not [ pkt for pkt in s.msgs if s.match_func( s.recv_msg, pkt) ]:
      # FIXME: s.idx does not mean anything here...
      s.error_msg = ("""
Test Sink {} received WRONG msg!
Received : {}
""".format( str(s), s.recv_msg ) )

    # Check performace regression
    elif s.perf_regr and s.cycle_count > s.arrival_time[ s.idx ]:
      s.error_msg = ("""
Test Sink {} received msg LATER than expected!
Message        : {}
Expected cycles: {}
Received at    : {}""".format( str(s), s.msgs[ s.idx ], s.arrival_time[ s.idx ], s.cycle_count ) )

    # No error
    else:
      s.idx += 1
      s.recv_called = True
      # Remove received pkt from list
      for pkt in s.msgs:
        if s.match_func( s.recv_msg, pkt ):
          s.msgs.remove( pkt )
          break

  def done( s ):
    return s.idx >= s.nmsgs

  # Line trace
  def line_trace( s ):
    return "{}".format( s.recv )

#-------------------------------------------------------------------------
# TestSinkRTL
#-------------------------------------------------------------------------

class TestNetSinkRTL( Component ):

  def construct( s,
    MsgType,
    msgs,
    initial_delay=0,
    interval_delay=0,
    arrival_time=None,
    match_func=lambda a, b : a.src == b.src and a.dst==b.dst and a.payload == b.payload,
  ):

    # Interface

    s.recv = RecvIfcRTL( MsgType )

    # Components

    s.sink    = TestNetSinkCL( MsgType, msgs, initial_delay, interval_delay,
                               arrival_time, match_func )
    s.adapter = RecvRTL2SendCL( MsgType )

    s.recv         //= s.adapter.recv
    s.adapter.send //= s.sink.recv

  def done( s ):
    return s.sink.done()

  # Line trace

  def line_trace( s ):
    return "{}".format( s.recv )



"""
==========================================================================
MeshRouterCL.py
==========================================================================
Cycle level implementation of MeshRouter.

Author : Yanghui Ou
  Date : May 16, 2019
"""
from pymtl3 import *
from router.InputUnitCL import InputUnitCL
from router.OutputUnitCL import OutputUnitCL
from router.Router import Router
from router.SwitchUnitCL import SwitchUnitCL

from .DORXMeshRouteUnitCL import DORXMeshRouteUnitCL


class MeshRouterCL( Router ):

  def construct( s,
    PacketType,
    PositionType,
    InputUnitType  = InputUnitCL,
    RouteUnitType  = DORXMeshRouteUnitCL,
    SwitchUnitType = SwitchUnitCL,
    OutputUnitType = OutputUnitCL,
  ):

    s.num_inports  = 5
    s.num_outports = 5

    # Interface
    s.pos  = InPort( PositionType )
    s.recv = [ CalleeIfcCL( Type=PacketType ) for _ in range( s.num_inports  ) ]
    s.send = [ CallerIfcCL( Type=PacketType ) for _ in range( s.num_outports ) ]

    # Components
    s.input_units  = [ InputUnitType( PacketType )
                      for _ in range( s.num_inports ) ]

    s.route_units  = [ RouteUnitType( PacketType, PositionType )
                      for i in range( s.num_inports ) ]

    s.switch_units = [ SwitchUnitType( PacketType, s.num_inports )
                      for _ in range( s.num_outports ) ]

    s.output_units = [ OutputUnitType( PacketType )
                      for _ in range( s.num_outports ) ]

    # Connection

    for i in range( s.num_inports ):
      s.recv[i]             //= s.input_units[i].recv
      s.input_units[i].give //= s.route_units[i].get
      s.pos                 //= s.route_units[i].pos

    for i in range( s.num_inports ):
      for j in range( s.num_outports ):
        s.route_units[i].give[j] //= s.switch_units[j].get[i]

    for j in range( s.num_outports ):
      s.switch_units[j].give //= s.output_units[j].get
      s.output_units[j].send //= s.send[j]

  def line_trace( s ):
    in_trace  = "|".join([ str(s.recv[i]) for i in range(5) ])
    out_trace = "|".join([ str(s.send[i]) for i in range(5) ])
    return "{}_({})_{}".format( in_trace, s.pos, out_trace )



'''
==========================================================================
connects_test.py
==========================================================================
unit tests for connect utils.

Author : Yanghui Ou
  Date : Feb 7, 2020
'''
from pymtl3 import *
from .connects import bitstruct_to_slices, connect_bitstruct

#-------------------------------------------------------------------------
# Bitstructs for unit tests
#-------------------------------------------------------------------------

@bitstruct
class Point:
  x : Bits8
  y : Bits4

@bitstruct
class Nested:
  z    : Bits6
  pt   : Point
  flag : Bits1

#-------------------------------------------------------------------------
# Components for unit tests
#-------------------------------------------------------------------------

class Bits2Bitstruct( Component ):
  def construct( s ):
    s.pt_bits      = InPort ( Bits12 )
    s.pt_bitstruct = OutPort( Point  )
    connect_bitstruct( s.pt_bits, s.pt_bitstruct )

class Bitstruct2Bits( Component ):
  def construct( s ):
    s.pt_bitstruct = InPort ( Point  )
    s.pt_bits      = OutPort( Bits12 )
    connect_bitstruct( s.pt_bitstruct, s.pt_bits )

class Bits2BitstructNested( Component ):
  def construct( s ):
    s.pt_bits      = InPort ( Bits19 )
    s.pt_bitstruct = OutPort( Nested )
    connect_bitstruct( s.pt_bits, s.pt_bitstruct )

class Bitstruct2BitsNested( Component ):
  def construct( s ):
    s.pt_bitstruct = InPort ( Nested )
    s.pt_bits      = OutPort( Bits19 )
    connect_bitstruct( s.pt_bitstruct, s.pt_bits )

#-------------------------------------------------------------------------
# Unit test fos helper funcion
#-------------------------------------------------------------------------

def test_bitstruct_to_slices():
  slices = bitstruct_to_slices( Point )
  assert slices[0] == slice(0, 4 )
  assert slices[1] == slice(4, 12)

  slices_nested = bitstruct_to_slices( Nested )
  assert slices_nested[0] == slice(0,  1 )
  assert slices_nested[1] == slice(1,  5 )
  assert slices_nested[2] == slice(5,  13)
  assert slices_nested[3] == slice(13, 19)

#-------------------------------------------------------------------------
# Unit test fos connect_bitstruct 
#-------------------------------------------------------------------------

def test_connet_bitstruct_simple():
  a = Bits2Bitstruct()
  a.elaborate()
  a.apply( SimulationPass() )
  a.pt_bits = b12( 0xeda )
  a.eval_combinational()
  assert a.pt_bitstruct.x == 0xed
  assert a.pt_bitstruct.y == 0xa

  b = Bitstruct2Bits()
  b.elaborate()
  b.apply( SimulationPass() )
  b.pt_bitstruct.x = b8( 0xed )
  b.pt_bitstruct.y = b4( 0xa  )
  b.eval_combinational()
  assert b.pt_bits == 0xeda

def test_connect_bitstruct_nested():
  a = Bits2BitstructNested()
  a.elaborate()
  a.apply( SimulationPass() )
  a.pt_bits = concat( b6(0x3f), b12(0xeda), b1(1) )
  a.eval_combinational()
  assert a.pt_bitstruct.z    == 0x3f
  assert a.pt_bitstruct.pt.x == 0xed
  assert a.pt_bitstruct.pt.y == 0xa
  assert a.pt_bitstruct.flag == 1

  b = Bitstruct2BitsNested()
  b.elaborate()
  b.apply( SimulationPass() )
  b.pt_bitstruct.z    = b6( 0x3f )
  b.pt_bitstruct.pt.x = b8( 0xed )
  b.pt_bitstruct.pt.y = b4( 0xa  )
  b.pt_bitstruct.flag = b1( 0x1  )
  b.eval_combinational()
  assert b.pt_bits == concat( b6(0x3f), b12(0xeda), b1(1) )




'''
==========================================================================
SwitchUnitNullRTL.py
==========================================================================
A null switch unit that is used for single input router.

Author : Yanghui Ou
  Date : Feb 21, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL

class SwitchUnitNullRTL( Component ):

  def construct( s, Type, num_inports=1 ):

    assert num_inports == 1, 'Null switch unit can only be used for single-input router!'

    # Interface

    s.get  = [ GetIfcRTL( Type ) for _ in range( num_inports )  ]
    s.hold = [ InPort( Bits1 ) for _ in range( num_inports ) ]
    s.give = GiveIfcRTL( Type )

    connect( s.give, s.get[0] )

  def line_trace( s ):
    in_trace  = '|'.join( [ str(p) for p in s.get ] )
    hold      = ''.join([ '^' if h else '.' for h in s.hold ])
    out_trace = f'{s.give}'
    return f'{in_trace}({hold}){out_trace}'



#=========================================================================
# MeshNetworkCL.py
#=========================================================================
# Mesh network cycle level implementation.
#
# Author : Yanghui Ou
#   Date : May 21, 2019

from channel.ChannelCL import ChannelCL
from pymtl3 import *
from pymtl3.stdlib.cl.queues import BypassQueueCL
from ocnlib.cl import BoundaryUnit

from .directions import *
from .MeshRouterCL import MeshRouterCL


class MeshNetworkCL( Component ):
  def construct( s, PacketType, PositionType,
                 ncols = 4, nrows = 4, chl_lat = 0 ):

    # Constants

    s.num_routers = ncols * nrows
    s.num_terminals = s.num_routers
    num_channels  = (nrows*(ncols-1)+ncols*(nrows-1)) * 2
    chl_lat       =  0

    # Interface

    s.recv = [ CalleeIfcCL( Type=PacketType ) for _ in range( s.num_terminals ) ]
    s.send = [ CallerIfcCL( Type=PacketType ) for _ in range( s.num_terminals ) ]

    # Components

    s.routers  = [ MeshRouterCL( PacketType, PositionType )
                 for i in range( s.num_routers ) ]

    s.channels = [ ChannelCL( PacketType, latency = chl_lat)
                 for _ in range( num_channels ) ]

    # Connect routers together in Mesh

    chl_id  = 0
    for i in range( s.num_routers ):
      if i // ncols > 0:
        s.routers[i].send[SOUTH] //= s.channels[chl_id].recv
        s.channels[chl_id].send  //= s.routers[i-ncols].recv[NORTH]
        chl_id += 1

      if i // ncols < nrows - 1:
        s.routers[i].send[NORTH] //= s.channels[chl_id].recv
        s.channels[chl_id].send  //= s.routers[i+ncols].recv[SOUTH]
        chl_id += 1

      if i % ncols > 0:
        s.routers[i].send[WEST] //= s.channels[chl_id].recv
        s.channels[chl_id].send //= s.routers[i-1].recv[EAST]
        chl_id += 1

      if i % ncols < ncols - 1:
        s.routers[i].send[EAST] //= s.channels[chl_id].recv
        s.channels[chl_id].send //= s.routers[i+1].recv[WEST]
        chl_id += 1

      # Connect the self port (with Network Interface)

      s.recv[i] //= s.routers[i].recv[SELF]
      s.send[i] //= s.routers[i].send[SELF]

      # Connect the unused ports to dummy methods
      # def dummy_rdy():
      #   return lambda : False

      # def dummy_method():
      #   ...

      # FIXME: this doesn't work!
      # if i // ncols == 0:
      #   s.routers[i].send[SOUTH].method.method = lambda s: None
      #   s.routers[i].send[SOUTH].rdy.method    = lambda s: False

      # if i // ncols == nrows - 1:
      #   s.routers[i].send[NORTH].method.method = lambda s: None
      #   s.routers[i].send[NORTH].rdy.method    = lambda s: False

      # if i % ncols == 0:
      #   s.routers[i].send[WEST].method.method = lambda s: None
      #   s.routers[i].send[WEST].rdy.method    = lambda s: False

      # if i % ncols == ncols - 1:
      #   s.routers[i].send[EAST].method.method = lambda s: None
      #   s.routers[i].send[EAST].rdy.method    = lambda s: False
      
      # Connet unused port to dummy queues
      s.dangling_q_n = [ BoundaryUnit( default_rdy=False ) for _ in range( ncols ) ]
      s.dangling_q_s = [ BoundaryUnit( default_rdy=False ) for _ in range( ncols ) ]
      s.dangling_q_w = [ BoundaryUnit( default_rdy=False ) for _ in range( nrows ) ]
      s.dangling_q_e = [ BoundaryUnit( default_rdy=False ) for _ in range( nrows ) ]

      if i // ncols == 0:
        s.routers[i].send[ SOUTH ] //= s.dangling_q_s[ i % ncols ].recv

      if i // ncols == nrows - 1:
        s.routers[i].send[ NORTH ] //= s.dangling_q_n[ i % ncols ].recv

      if i % ncols == 0:
        s.routers[i].send[ WEST ] //= s.dangling_q_w[ i // ncols ].recv

      if i % ncols == ncols - 1:
        s.routers[i].send[ EAST ] //= s.dangling_q_e[ i // ncols ].recv

    # Set the position of each router
    for y in range( nrows ):
      for x in range( ncols ):
        idx = y * ncols + x
        s.routers[idx].pos //= PositionType( x, y )

  def line_trace( s ):
    return "|".join( [ s.routers[i].line_trace() for i in range(s.num_routers) ] )

#-------------------------------------------------------------------------
# WrappedMeshNetCL
#-------------------------------------------------------------------------
# A wrapped mesh net exposing only callee ports.

class WrappedMeshNetCL( Component ):

  def construct( s,
    PacketType,
    PositionType,
    ncols=4,
    nrows=4,
    chl_lat = 0
  ):
    s.nterminals = nrows * ncols
    s.recv = [ CalleeIfcCL( PacketType )  for _ in range( s.nterminals ) ]
    s.give = [ CalleeIfcCL( PacketType )  for _ in range( s.nterminals ) ]
    s.net = MeshNetworkCL( PacketType, PositionType, ncols, nrows, chl_lat )
    s.out_q = [ BypassQueueCL( num_entries=1 ) for _ in range( s.nterminals ) ]
    for i in range( s.nterminals ):
      s.recv[i]      //= s.net.recv[i]
      s.net.send[i]  //= s.out_q[i].enq
      s.out_q[i].deq //= s.give[i]

  def line_trace( s ):
    return s.net.line_trace()



'''
==========================================================================
MflitPacket.py
==========================================================================
A **non translatable** data type that helps testing components that use
multi-flit (single-phit flit) packets.

Author : Yanghui Ou
  Date : Jan 31, 2019
'''
from pymtl3 import mk_bits
from ..utils import get_nbits, to_bitstruct
from pymtl3.datatypes.bitstructs import (
  is_bitstruct_class, 
  _FIELDS as bitstruct_fields,
)

#-------------------------------------------------------------------------
# _get_payload_length
#-------------------------------------------------------------------------

def _get_payload_length( Format, header_flit, field_name='plen' ):
  header = to_bitstruct( header_flit, Format )
  return getattr( header, field_name ).uint()

#-------------------------------------------------------------------------
# MflitPacket
#-------------------------------------------------------------------------

class MflitPacket:

  def __init__( self, Format, flits=[], plen_field_name='plen' ):

    self.Format     = Format
    self.plen_fname = plen_field_name 
    self.PhitType   = mk_bits( get_nbits( Format ) )
    self.add_lock   = False
    self.pop_lock   = False
    self.nflits     = 0
    self.flit_idx   = 0
    self.flits      = [ self.PhitType(f) for f in flits ]

    # Check
    if self.flits:
      self._get_nflits()
      assert self.nflits == len( self.flits )

  def add( self, flit ):
    assert not self.pop_lock, "Packet locked by pop, cannot add any more!"

    # Adding header flit
    if self.empty():
      self.flits.append( flit )
      self._get_nflits()
    else:
      assert not self.full(), "Packet is already full" 
      self.flits.append( flit )

    self.add_lock = True

  def pop( self ):
    assert not self.add_lock, "Packet locked by add, cannot pop any more!"
    assert self.flit_idx < self.nflits, "Already reached the last flit of the packet!" 
    # Copy the current flit
    cur_flit = self.PhitType( self.flits[ self.flit_idx ] )

    self.flit_idx += 1
    self.pop_lock = True
    return cur_flit

  def full( self ):
    return self.nflits > 0 and self.nflits == len( self.flits ) and self.flit_idx == 0

  def empty( self ):
    return len( self.flits ) == 0 or self.flit_idx == self.nflits

  def _get_nflits( self ):
    nflits = _get_payload_length( self.Format, self.flits[0], self.plen_fname ) + 1
    self.nflits = nflits



"""
==========================================================================
InputOutputUnitCredit_test.py
==========================================================================
Composition test for input/output unit with credit based interfaces.

Author : Yanghui Ou
  Date : June 22, 2019
"""
from ocnlib.ifcs.CreditIfc import (CreditRecvRTL2SendRTL,
                                      RecvRTL2CreditSendRTL)
from ocnlib.ifcs.packets import mk_generic_pkt
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkRTL
from pymtl3 import *
from pymtl3.stdlib.rtl.queues import BypassQueueRTL
from pymtl3.stdlib.test.test_srcs import TestSrcRTL

from ..InputUnitCreditRTL import InputUnitCreditRTL
from ..OutputUnitCreditRTL import OutputUnitCreditRTL
from ..SwitchUnitRTL import SwitchUnitRTL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):
  def construct( s, Type, src_msgs, sink_msgs, vc=2, credit_line=2 ):

    s.src   = TestSrcRTL( Type, src_msgs )
    s.src_q = BypassQueueRTL( Type, num_entries=1 )
    s.output_unit = OutputUnitCreditRTL( Type, vc=vc, credit_line=credit_line )
    s.input_unit  = InputUnitCreditRTL( Type, vc=vc, credit_line=credit_line )
    s.switch_unit = SwitchUnitRTL( Type, num_inports=vc )
    s.sink = TestNetSinkRTL( Type, sink_msgs )

    s.src.send             //= s.src_q.enq
    s.src_q.deq            //= s.output_unit.get
    s.output_unit.send     //= s.input_unit.recv
    s.switch_unit.give.ret //= s.sink.recv.msg
    for i in range( vc ):
      s.input_unit.give[i] //= s.switch_unit.get[i]

    @s.update
    def up_sink_recv():
      s.sink.recv.en = s.switch_unit.give.rdy & s.sink.recv.rdy
      s.switch_unit.give.en = s.switch_unit.give.rdy & s.sink.recv.rdy

  def line_trace( s ):
    return "{} >>> {} >>> {} >>> {} >>> {}".format(
      s.src.line_trace(),
      s.output_unit.line_trace(),
      s.input_unit.line_trace(),
      s.switch_unit.line_trace(),
      s.sink.line_trace()
    )

  def done( s ):
    return s.src.done() and s.sink.done()

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

def test_simple():
  Pkt = mk_generic_pkt( vc=2 )
  msgs = [
    Pkt( 0, 1, 0x04, 0, 0xdeadbabe ),
    Pkt( 0, 2, 0x02, 1, 0xfaceb00c ),
    Pkt( 0, 3, 0x03, 0, 0xdeadface ),
  ]
  th = TestHarness( Pkt, msgs, msgs )
  run_sim( th )

def test_backpresure():
  Pkt = mk_generic_pkt( vc=2 )
  msgs = [
      # src dst opq vc_id payload
    Pkt( 0, 1, 0x04, 0, 0xdeadbabe ),
    Pkt( 0, 2, 0x02, 1, 0xfaceb00c ),
    Pkt( 0, 3, 0x03, 0, 0xdeadface ),
    Pkt( 0, 3, 0x03, 1, 0xdeadface ),
    Pkt( 0, 3, 0x03, 1, 0xdeadface ),
    Pkt( 0, 3, 0x03, 0, 0xdeadface ),
    Pkt( 0, 3, 0x03, 0, 0xdeadface ),
  ]
  th = TestHarness( Pkt, msgs, msgs )
  th.set_param( "top.sink.construct", initial_delay=20 )
  run_sim( th )



'''
==========================================================================
commons.py
==========================================================================
Common utility functions.

Author : Yanghui Ou
  Date : Feb 9, 2020
'''
from functools import reduce
from pymtl3 import *
from pymtl3.datatypes.bitstructs import(
  is_bitstruct_class,
  is_bitstruct_inst,
  _FIELDS as bitstruct_fields,
)

#-------------------------------------------------------------------------
# get_nbits
#-------------------------------------------------------------------------
# TODO: support instance too

def get_nbits( cls ):
  if issubclass( cls, Bits ):
    return cls.nbits

  else:
    assert is_bitstruct_class( cls )
    fields_dict = getattr( cls, bitstruct_fields )
    total_nbits = 0
    for _, ftype in fields_dict.items():
      total_nbits +=  get_nbits( ftype )
    return total_nbits

#-------------------------------------------------------------------------
# get_plen_type
#-------------------------------------------------------------------------
# Returns the type of feild payload length from the header format.
# FIXME: hacky.

def get_plen_type( cls, plen_field_name='plen' ):
  assert is_bitstruct_class( cls )
  fields_dict = getattr( cls, bitstruct_fields )

  if not plen_field_name in fields_dict:
    raise AssertionError( f'{cls.__qualname__} does not have field {plen_field_name}!' )

  return fields_dict[ plen_field_name ]

#-------------------------------------------------------------------------
# get_field_type( cls ):
#-------------------------------------------------------------------------

def get_field_type( cls, field_name ):
  assert is_bitstruct_class( cls )
  fields_dict = getattr( cls, bitstruct_fields )

  if not field_name in fields_dict:
    raise AssertionError( f'{cls.__qualname__} does not have field {field_name}!' )

  return fields_dict[ field_name ]

#-------------------------------------------------------------------------
# bitstruct_to_slice_h
#-------------------------------------------------------------------------
# Recursive helper function

def _bitstruct_to_slices_h( cls, slices ):
  if issubclass( cls, Bits ):
    start = 0 if not slices else slices[-1].stop
    stop  = start + cls.nbits
    slices.append( slice( start, stop ) )

  else:
    assert is_bitstruct_class( cls )
    fields_dict = getattr( cls, bitstruct_fields )
    fields_lst  = list( fields_dict.items() )
    # Use reverse instead of [::-1] because it is faster
    fields_lst.reverse()
    for _, ftype in fields_lst:
      _bitstruct_to_slices_h( ftype, slices )

#-------------------------------------------------------------------------
# bitstruct_to_slice
#-------------------------------------------------------------------------
# Converts a bitstruct to a list of slices

def bitstruct_to_slices( cls ):
  slices = []
  _bitstruct_to_slices_h( cls, slices )
  return slices

#-------------------------------------------------------------------------
# to_bits
#-------------------------------------------------------------------------
# Converts a bitstruct object to bits

def to_bits( obj ):
  if isinstance( obj, Bits ):
    # FIXME: call clone here once pymtl3 is updated
    return obj.__class__( obj )

  else:
    assert is_bitstruct_inst( obj )
    fields_dict = getattr( obj, bitstruct_fields )

    acc = []
    for fname, ftype in fields_dict.items():
      bits = to_bits( getattr( obj, fname ) )
      acc.append( bits )

    return reduce( lambda a, b: concat( a, b ), acc )

#-------------------------------------------------------------------------
# _to_bitstruct_h
#-------------------------------------------------------------------------
# Recursive helper function for to_bitstruct.
# NOTE: I am being very conservative here to copy every field.

def _to_bitstruct_h( field_type, bits_obj, slices_stack ):
  if issubclass( field_type, Bits ):
    cur_slice = slices_stack.pop()
    return field_type( bits_obj[ cur_slice ] )

  else:
    assert is_bitstruct_class( field_type )
    ret = field_type()
    fields_dict = getattr( field_type, bitstruct_fields )
    for fname, ftype in fields_dict.items():
      field_value = _to_bitstruct_h( ftype, bits_obj, slices_stack )
      setattr( ret, fname, field_value )
    return ret

#-------------------------------------------------------------------------
# to_bitstruct
#-------------------------------------------------------------------------
# Converts a bit or bitstruct object to bitstruct
# TODO: more informative error message.

def to_bitstruct( obj, BitstructType ):
  bits_obj = to_bits( obj )
  assert isinstance( bits_obj, Bits )

  slices_stack = bitstruct_to_slices( BitstructType )
  assert bits_obj.nbits == slices_stack[-1].stop, "Bit width mismatch!"
  ret = _to_bitstruct_h( BitstructType, bits_obj, slices_stack )
  assert not slices_stack # Make sure every field is assigned
  return ret

#-------------------------------------------------------------------------
# run_sim
#-------------------------------------------------------------------------
# A generic run_sim function

def run_sim( th, max_cycles=1000, translation='', 
             dut_name='dut', vl_trace=False ):

  th.elaborate()

  if translation == 'verilog':
    from pymtl3.passes.backends.verilog import TranslationImportPass
    from pymtl3.passes.backends.verilog import VerilatorImportConfigs
    getattr( th, dut_name ).verilog_translate_import = True
    getattr( th, dut_name ).config_verilog_import = VerilatorImportConfigs(
      vl_trace = vl_trace,
    )

  elif translation == 'yosys':
    from pymtl3.passes.backends.yosys import TranslationImportPass
    from pymtl3.passes.backends.yosys import VerilatorImportConfigs
    getattr( th, dut_name ).yosys_translate_import = True
    getattr( th, dut_name ).config_yosys_import = VerilatorImportConfigs(
      vl_trace = vl_trace,
    )

  elif translation:
    assert False, f'Invalid translation backend {translation}!'

  if translation:
    th = TranslationImportPass()( th )
    th.elaborate()

  th.apply( SimulationPass() )
  th.sim_reset()

  # Run simulation
  print()
  th.print_line_trace()
  while not th.done() and th.simulated_cycles < max_cycles:
    th.tick()
    th.print_line_trace()

  # Check timeout
  assert th.simulated_cycles < max_cycles
  th.tick()
  th.tick()
  th.tick()



#=========================================================================
# MeshRouterRTL_test.py
#=========================================================================
# Test for RouterRTL
#
# Author : Cheng Tan, Yanghui Ou
#   Date : Mar 10, 2019

import hypothesis
from hypothesis import strategies as st

from meshnet.DORXMeshRouteUnitRTL import DORXMeshRouteUnitRTL
from meshnet.DORYMeshRouteUnitRTL import DORYMeshRouteUnitRTL
from meshnet.MeshRouterRTL import MeshRouterRTL
from ocnlib.ifcs.packets import mk_mesh_pkt
from ocnlib.ifcs.positions import mk_mesh_pos
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkRTL
from pymtl3 import *
from pymtl3.stdlib.test.test_srcs import TestSrcRTL
from router.InputUnitRTL import InputUnitRTL
from router.OutputUnitRTL import OutputUnitRTL
from router.SwitchUnitRTL import SwitchUnitRTL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s,
                 MsgType       = None,
                 ncols      = 2,
                 nrows       = 2 ,
                 pos_x         = 0,
                 pos_y         = 0,
                 src_msgs      = [],
                 sink_msgs     = [],
                 src_initial   = 0,
                 src_interval  = 0,
                 sink_initial  = 0,
                 sink_interval = 0,
                 arrival_time  =[None, None, None, None, None]
               ):

    MeshPos = mk_mesh_pos( ncols, nrows )
    s.dut = MeshRouterRTL( MsgType, MeshPos, InputUnitType = InputUnitRTL,
        RouteUnitType = DORYMeshRouteUnitRTL )
    match_func = lambda a, b : a.payload == b.payload

    s.srcs  = [ TestSrcRTL    ( MsgType, src_msgs[i],  src_initial,  src_interval  )
                for i in range  ( s.dut.num_inports ) ]
    s.sinks = [ TestNetSinkRTL( MsgType, sink_msgs[i], sink_initial,
                match_func=match_func ) for i in range ( s.dut.num_outports ) ]

    # Connections

    for i in range ( s.dut.num_outports ):
      s.srcs[i].send //= s.dut.recv[i]
      s.dut.send[i]  //= s.sinks[i].recv

    @s.update
    def up_pos():
      s.dut.pos = MeshPos( pos_x, pos_y )

  def done( s ):
    srcs_done = 1
    sinks_done = 1

    for x in s.srcs:
      if x.done() == 0:
        srcs_done = 0

    for x in s.sinks:
      if x.done() == 0:
        sinks_done = 0
    return srcs_done and sinks_done

  def line_trace( s ):
    return "{}".format( s.dut.line_trace() )

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

#              x,y,pl,dir
test_msgs = [[(0,0,11,1),(0,0,12,1),(0,1,13,2),(2,1,14,3),(0,0,15,1)],
             [(0,0,21,1),(0,2,22,0),(0,1,23,2),(2,1,24,3),(2,1,25,3)],
             [(0,2,31,0),(0,0,32,1),(0,1,33,2),(1,1,34,4),(1,1,35,4)]
            ]
result_msgs = [[],[],[],[],[]]

def test_normal_simple():
  src_packets = [[],[],[],[],[]]
  for item in test_msgs:
    for i in range( len( item ) ):
      (dst_x,dst_y,payload,dir_out) = item[i]
      PacketType = mk_mesh_pkt (4, 4)
      pkt = PacketType (0, 0, dst_x, dst_y, 1, payload)
      src_packets[i].append( pkt )
      result_msgs[dir_out].append( pkt )

  th = TestHarness( PacketType, 4, 4, 1, 1, src_packets, result_msgs, 0, 0, 0, 0 )
  run_sim( th )

def test_self_simple():
  PacketType = mk_mesh_pkt(4, 4)
  pkt = PacketType( 0, 0, 0, 0, 0, 0xdead )
  src_pkts  = [ [], [], [], [], [pkt] ]
  sink_pkts = [ [], [], [], [], [pkt] ]
  th = TestHarness( PacketType, 4, 4, 0, 0, src_pkts, sink_pkts )
  run_sim( th )

def test_h1():
  pos_x, pos_y, ncols, nrows = 0, 0, 2, 2
  PacketType = mk_mesh_pkt( ncols, nrows )
  pkt0 = PacketType( 0, 0, 0, 1, 0, 0xdead )
  src_pkts  = [ [],     [], [], [], [pkt0] ]
  sink_pkts = [ [pkt0], [], [], [], []     ]
  th = TestHarness(
    PacketType, ncols, nrows, pos_x, pos_y,
    src_pkts, sink_pkts
  )
  th.set_param(
    "top.dut.construct",
    RouteUnitType = DORYMeshRouteUnitRTL
  )
  run_sim( th )

def test_h2():
  pos_x, pos_y, ncols, nrows = 0, 0, 2, 2
  PacketType = mk_mesh_pkt( ncols, nrows )
  pkt0 = PacketType( 0, 0, 1, 0, 0, 0xdead )
  pkt1 = PacketType( 0, 1, 1, 0, 1, 0xbeef )
  pkt2 = PacketType( 0, 1, 1, 0, 2, 0xcafe )
              # N             S   W   E                   self
  src_pkts  = [ [pkt1, pkt2], [], [], [],                 [pkt0] ]
  sink_pkts = [ [],           [], [], [pkt1, pkt0, pkt2], []     ]
  th = TestHarness(
    PacketType, ncols, nrows, pos_x, pos_y,
    src_pkts, sink_pkts
  )
  th.set_param(
    "top.dut.construct",
    RouteUnitType = DORYMeshRouteUnitRTL
  )
  run_sim( th, 10 )

def test_h3():
  pos_x, pos_y, ncols, nrows = 0, 1, 2, 2
  PacketType = mk_mesh_pkt( ncols, nrows )
  pkt0 = PacketType( 0, 1, 0, 0, 0, 0xdead )
              # N   S   W   E   self
  src_pkts  = [ [], [], [], [], [pkt0] ]
  sink_pkts = [ [], [pkt0], [], [], [] ]
  th = TestHarness(
    PacketType, ncols, nrows, pos_x, pos_y,
    src_pkts, sink_pkts
  )
  th.set_param(
    "top.dut.construct",
    RouteUnitType = DORYMeshRouteUnitRTL
  )
  run_sim( th, 10 )



"""
==========================================================================
MeshRouteFL.py
==========================================================================
Functional level implementation of mesh router.

Author : Yanghui Ou
  Date : July 3, 2019
"""
from pymtl3 import *
from ocnlib.utils import to_bitstruct

from .directions import *


#-------------------------------------------------------------------------
# MeshRouterFL
#-------------------------------------------------------------------------
# Mesh router that supports single flit packet.

class MeshRouterFL:

  def __init__( s, pos_x, pos_y, dimension='y' ):
    s.pos_x = pos_x
    s.pos_y = pos_y
    s.dimension = dimension

  def arrange_src_pkts( s, lst ):
    src_pkts = [ [] for _ in range(5) ]
    if s.dimension == 'y':
      for pkt in lst:
        if pkt.src_x == s.pos_x and pkt.src_y == s.pos_y:
          src_pkts[SELF].append( pkt )
        elif pkt.src_x == s.pos_x:
          if pkt.src_y < s.pos_y:
            src_pkts[SOUTH].append( pkt )
          else:
            src_pkts[NORTH].append( pkt )
        elif pkt.src_x < s.pos_x:
          src_pkts[WEST].append( pkt )
        else:
          src_pkts[EAST].append( pkt )

    else: # s.dimension == 'x'
      for pkt in lst:
        if pkt.src_x == s.pos_x and pkt.src_y == s.pos_y:
          src_pkts[SELF].append( pkt )
        elif pkt.src_y == s.pos_y:
          if pkt.src_x < s.pos_x:
            src_pkts[WEST].append( pkt )
          else:
            src_pkts[NORTH].append( pkt )
        elif pkt.src_y < s.pos_y:
          src_pkts[SOUTH].append( pkt )
        else:
          src_pkts[NORTH].append( pkt )

    return src_pkts

  def route( s, src_pkts ):
    assert len( src_pkts ) == 5
    dst_pkts = [ [] for _ in range(5) ]

    if s.dimension == 'y':
      for pkts in src_pkts:
        for pkt in pkts:
          dst = (
            SELF  if pkt.dst_x == s.pos_x and pkt.dst_y == s.pos_y else
            NORTH if pkt.dst_y > s.pos_y else
            SOUTH if pkt.dst_y < s.pos_y else
            EAST  if pkt.dst_x > s.pos_x else
            WEST
          )
          dst_pkts[ dst ].append( pkt )

    else: # s.dimension == 'x'
      for pkts in src_pkts:
        for pkt in pkts:
          dst = (
            SELF  if pkt.dst_x == s.pos_x and pkt.dst_y == s.pos_y else
            EAST  if pkt.dst_x > s.pos_x else
            WEST  if pkt.dst_x < s.pos_x else
            NORTH if pkt.dst_y > s.pos_y else
            SOUTH
          )
          dst_pkts[ dst ].append( pkt )

    return dst_pkts

#-------------------------------------------------------------------------
# MeshRouterMflitFL
#-------------------------------------------------------------------------
# Mesh router that supports single flit packet.

class MeshRouterMflitFL:

  def __init__( s, Header, pos_x, pos_y, first_dimension='x' ):
    s.Header          = Header
    s.pos_x           = pos_x
    s.pos_y           = pos_y
    s.first_dimension = first_dimension

  def arrange_src_pkts( s, lst ):
    src_pkts = [ [] for _ in range(5) ]
    if s.first_dimension == 'y':
      for pkt in lst:
        header = to_bitstruct( pkt.flits[0], s.Header )
        if header.src_x == s.pos_x and header.src_y == s.pos_y:
          src_pkts[ SELF ].append( pkt )
        elif header.src_x == s.pos_x:
          if header.src_y < s.pos_y:
            src_pkts[ SOUTH ].append( pkt )
          else:
            src_pkts[ NORTH ].append( pkt )
        elif header.src_x < s.pos_x:
          src_pkts[ WEST ].append( pkt )
        else:
          src_pkts[ EAST ].append( pkt )

    elif s.first_dimension == 'x':
      for pkt in lst:
        header = to_bitstruct( pkt.flits[0], s.Header )
        if header.src_x == s.pos_x and header.src_y == s.pos_y:
          src_pkts[ SELF ].append( pkt )
        elif header.src_y == s.pos_y:
          if header.src_x < s.pos_x:
            src_pkts[ WEST ].append( pkt )
          else:
            src_pkts[ NORTH ].append( pkt )
        elif header.src_y < s.pos_y:
          src_pkts[ SOUTH ].append( pkt )
        else:
          src_pkts[ NORTH ].append( pkt )
    else:
      assert False

    return src_pkts

  def route( s, src_pkts ):
    assert len( src_pkts ) == 5
    dst_pkts = [ [] for _ in range(5) ]

    if s.first_dimension == 'y':
      for pkts in src_pkts:
        for pkt in pkts:
          header = to_bitstruct( pkt.flits[0], s.Header )
          dst = (
            SELF  if header.dst_x == s.pos_x and pkt.dst_y == s.pos_y else
            NORTH if header.dst_y > s.pos_y else
            SOUTH if header.dst_y < s.pos_y else
            EAST  if header.dst_x > s.pos_x else
            WEST
          )
          dst_pkts[ dst ].append( pkt )

    elif s.first_dimension == 'x':
      for pkts in src_pkts:
        for pkt in pkts:
          header = to_bitstruct( pkt.flits[0], s.Header )
          dst = (
            SELF  if header.dst_x == s.pos_x and header.dst_y == s.pos_y else
            EAST  if header.dst_x > s.pos_x else
            WEST  if header.dst_x < s.pos_x else
            NORTH if header.dst_y > s.pos_y else
            SOUTH
          )
          dst_pkts[ dst ].append( pkt )

    else:
      assert False

    return dst_pkts



"""
==========================================================================
packets.py
==========================================================================
Collection of packets definition.

Convention: The fields/constructor arguments should appear in the order
            of [ <id related>, opaque_nbits, vc, payload_nbits ]

Author : Yanghui Ou, Shunning Jiang
  Date : Oct 26, 2019
"""
from pymtl3 import *

#=========================================================================
# Generic packet
#=========================================================================

def mk_generic_pkt( nrouters=4, opaque_nbits=8, vc=2, payload_nbits=16,
                    prefix="GenericPacket" ):

  IdType = mk_bits( clog2( nrouters ) )
  OpqType = mk_bits( opaque_nbits )
  PayloadType = mk_bits( payload_nbits )

  new_name = f"{prefix}_{nrouters}_{vc}_{opaque_nbits}_{payload_nbits}"

  if vc > 1:
    VcIdType = mk_bits( clog2( vc ) )

    def str_func( s ):
      return f"{s.src}>{s.dst}:{s.opaque}:{s.vc_id}:{s.payload}"

    return mk_bitstruct( new_name, {
        'src':     IdType,
        'dst':     IdType,
        'opaque':  OpqType,
        'vc_id':   VcIdType,
        'payload': PayloadType,
      },
      namespace = { '__str__': str_func }
    )

  else:
    def str_func( s ):
      return f"{s.src}>{s.dst}:{s.opaque}:{s.payload}"

    return mk_bitstruct( new_name, {
        'src':     IdType,
        'dst':     IdType,
        'opaque':  OpqType,
        'payload': PayloadType,
      },
      namespace = { '__str__': str_func }
    )

#=========================================================================
# ring packet
#=========================================================================

def mk_ring_pkt( nrouters=4, opaque_nbits=8, vc=2, payload_nbits=32 ):
  return mk_generic_pkt( nrouters, opaque_nbits, vc, payload_nbits, "RingPacket" )

#=========================================================================
# mesh packet
#=========================================================================

def mk_mesh_pkt( ncols=2, nrows=2,
                 opaque_nbits=8, vc=1, payload_nbits=32 ):
  assert ncols > 0 and nrows > 0

  XType       = mk_bits(clog2( ncols )) if ncols != 1 else Bits1
  YType       = mk_bits(clog2( nrows)) if nrows!= 1 else Bits1
  OpqType     = mk_bits(opaque_nbits)
  PayloadType = mk_bits(payload_nbits)

  new_name = f"MeshPacket_{ncols}x{nrows}_{vc}_{opaque_nbits}_{payload_nbits}"

  if vc > 1:
    VcIdType = mk_bits( clog2( vc ) )

    def str_func( s ):
      return f"({s.src_x},{s.src_y})>({s.dst_x},{s.dst_y}):{s.opaque}:{s.vc_id}:{s.payload}"

    return mk_bitstruct( new_name, {
        'src_x':   XType,
        'src_y':   YType,
        'dst_x':   XType,
        'dst_y':   YType,
        'opaque':  OpqType,
        'vc_id':   VcIdType,
        'payload': PayloadType,
      },
      namespace = { '__str__': str_func },
    )
  else:
    def str_func( s ):
      return f"({s.src_x},{s.src_y})>({s.dst_x},{s.dst_y}):{s.opaque}:{s.payload}"

    return mk_bitstruct( new_name, {
        'src_x':   XType,
        'src_y':   YType,
        'dst_x':   XType,
        'dst_y':   YType,
        'opaque':  OpqType,
        'payload': PayloadType,
      },
      namespace = { '__str__': str_func }
    )

#=========================================================================
# cmesh packet
#=========================================================================

def mk_cmesh_pkt( ncols=2, nrows=2,
                  inports=8, outports=8,
                  opaque_nbits=8, vc=1, payload_nbits=32 ):

  XType = mk_bits(clog2(ncols))  if ncols  != 1 else Bits1
  YType = mk_bits(clog2(nrows)) if nrows != 1 else Bits1
  TType = mk_bits(clog2(outports-4))  if outports > 5 else Bits1
  OpqType = mk_bits( opaque_nbits )
  PayloadType = mk_bits( payload_nbits )

  new_name = f"CMeshPacket_{ncols}x{nrows}_{inports}x{outports}_" \
             f"{opaque_nbits}_{vc}_{payload_nbits}"

  if vc > 1:
    VcIdType = mk_bits( clog2( vc ) )
    return mk_bitstruct( new_name, {
      'src_x':   XType,
      'src_y':   YType,
      'dst_x':   XType,
      'dst_y':   YType,
      'dst_ter': TType,
      'opaque':  OpqType,
      'vc_id':   VcIdType,
      'payload': PayloadType,
    })
  else:
    return mk_bitstruct( new_name, {
      'src_x':   XType,
      'src_y':   YType,
      'dst_x':   XType,
      'dst_y':   YType,
      'dst_ter': TType,
      'opaque':  OpqType,
      'payload': PayloadType,
    })

#=========================================================================
# Butterfly packet
#=========================================================================

def mk_bfly_pkt( k_ary=2, n_fly=2, opaque_nbits=8, vc=0, payload_nbits=32 ):
  IdType   = mk_bits( clog2( k_ary ** n_fly ) )

  assert k_ary > 1

  DstType = mk_bits( clog2( k_ary ) * n_fly )

  OpqType = mk_bits( opaque_nbits )
  PayloadType = mk_bits( payload_nbits )

  new_name = f"BflyPacket_{k_ary}_{n_fly}_{vc}_{opaque_nbits}_{payload_nbits}"

  if vc > 1:
    VcIdType = mk_bits( clog2(vc) )

    return mk_bitstruct( new_name, {
      'src':     IdType,
      'dst':     DstType,
      'opaque':  OpqType,
      'vc_id':   VcIdType,
      'payload': PayloadType,
    })
  else:
    return mk_bitstruct( new_name, {
      'src':     IdType,
      'dst':     DstType,
      'opaque':  OpqType,
      'payload': PayloadType,
    })

#=========================================================================
# ring packet with timestamp
#=========================================================================

def mk_ring_pkt_timestamp( nrouters=4, opaque_nbits=8, vc=2, payload_nbits=32, max_time=10 ):

  IdType = mk_bits( clog2( nrouters ) )
  OpqType = mk_bits( opaque_nbits )
  PayloadType = mk_bits( payload_nbits )
  TimestampType = mk_bits( clog2(max_time + 1) )
  new_name = f"RingPacketTimestamp_{nrouters}_{opaque_nbits}_{vc}_{payload_nbits}_{max_time}"

  if vc > 1:
    VcIdType = mk_bits( clog2( vc ) )

    def str_func( s ):
      return f"{s.src}>{s.dst}:{s.opaque}:{s.vc_id}:{s.payload}:{s.timestamp}"

    return mk_bitstruct( new_name, {
        'src':       IdType,
        'dst':       IdType,
        'opaque':    OpqType,
        'vc_id':     VcIdType,
        'payload':   PayloadType,
        'timestamp': TimestampType,
      },
      namespace = { '__str__': str_func }
    )
  else:
    def str_func( s ):
      return f"{s.src}>{s.dst}:{s.opaque}:{s.payload}:{s.timestamp}"

    return mk_bitstruct( new_name, {
        'src':       IdType,
        'dst':       IdType,
        'opaque':    OpqType,
        'payload':   PayloadType,
        'timestamp': TimestampType,
      },
      namespace = { '__str__': str_func }
    )

#=========================================================================
# mesh packet with timestamp
#=========================================================================

def mk_mesh_pkt_timestamp( ncols=2, nrows=2,
                           opaque_nbits=8, vc=1, payload_nbits=32,
                           max_time=10 ):

  XType = mk_bits( clog2( ncols ) )
  YType = mk_bits( clog2( nrows  ) )
  OpqType = mk_bits( opaque_nbits )
  PayloadType = mk_bits( payload_nbits )
  TimestampType = mk_bits( clog2(max_time + 1) )
  new_name = f"MeshPacketTimestamp_{ncols}x{nrows}_{opaque_nbits}"\
             f"_{vc}_{payload_nbits}_{max_time}"

  if vc > 1:
    VcIdType = mk_bits( clog2( vc ) )

    def str_func( s ):
      return f"({s.src_x},{s.src_y})>({s.dst_x},{s.dst_y}):{s.opaque}:{s.vc_id}:{s.payload}:{s.timestamp}"

    return mk_bitstruct( new_name, {
        'src_x':     XType,
        'src_y':     YType,
        'dst_x':     XType,
        'dst_y':     YType,
        'opaque':    OpqType,
        'vc_id':     VcIdType,
        'payload':   PayloadType,
        'timestamp': TimestampType,
      },
      namespace = { '__str__': str_func },
    )

  else:
    def str_func( s ):
      return f"({s.src_x},{s.src_y})>({s.dst_x},{s.dst_y}):{s.opaque}:{s.payload}:{s.timestamp}"

    return mk_bitstruct( new_name, {
        'src_x':     XType,
        'src_y':     YType,
        'dst_x':     XType,
        'dst_y':     YType,
        'opaque':    OpqType,
        'payload':   PayloadType,
        'timestamp': TimestampType,
      },
      namespace = { '__str__': str_func },
    )

#=========================================================================
# cmesh packet with timestamp
#=========================================================================

def mk_cmesh_pkt_timestamp( ncols=2, nrows=2, inports=8, outports=8,
    opaque_nbits=8, vc=1, payload_nbits=32, max_time=10 ):

  XType = mk_bits( clog2( ncols ) )
  YType = mk_bits( clog2( nrows  ) )
  TType = mk_bits(clog2( outports - 4 )) if outports > 5 else mk_bits(1)
  OpqType = mk_bits( opaque_nbits )
  PayloadType = mk_bits( payload_nbits )
  TimestampType = mk_bits( clog2(max_time + 1) )
  new_name = f"MeshPacketTimestamp_{ncols}x{nrows}_{inports}x{outports}"\
             f"_{vc}_{opaque_nbits}_{payload_nbits}_{max_time}"

  if vc > 1:
    VcIdType = mk_bits( clog2( vc ) )
    return mk_bitstruct( new_name, {
      'src_x':     XType,
      'src_y':     YType,
      'dst_x':     XType,
      'dst_y':     YType,
      'dst_ter':   TType,
      'opaque':    OpqType,
      'vc_id':     VcIdType,
      'payload':   PayloadType,
      'timestamp': TimestampType,
    })
  else:
    return mk_bitstruct( new_name, {
      'src_x':     XType,
      'src_y':     YType,
      'dst_x':     XType,
      'dst_y':     YType,
      'dst_ter':   TType,
      'opaque':    OpqType,
      'payload':   PayloadType,
      'timestamp': TimestampType,
    })

#=========================================================================
# Butterfly packet with timestamp
#=========================================================================

def mk_bfly_pkt_timestamp( k_ary=2, n_fly=2,
                           opaque_nbits=8, vc=0, payload_nbits=32,
                           max_time=10 ):

  IdType   = mk_bits( clog2( k_ary ** n_fly ) )
#  if k_ary == 1:
#    KaryType = Bits1
#  else:
#    KaryType = mk_bits( clog2( k_ary ) )
#  if n_fly == 1:
#    NflyType = Bits1
#  else:
#    NflyType = mk_bits( clog2( n_fly ) )
#  if k_ary ** ( n_fly - 1 ) == 1:
#    RrowType = Bits1
#  else:
#    RrowType = mk_bits( clog2( k_ary ** ( n_fly - 1 ) ) )
  DstType = mk_bits( clog2( k_ary ) * n_fly )
#    DstType = mk_bits( clog2( k_ary ** ( n_fly - 1 ) ) * n_fly )
  OpqType = mk_bits( opaque_nbits )
  PayloadType = mk_bits( payload_nbits )
  TimestampType = mk_bits( clog2(max_time + 1) )

  new_name = f"BflyPacketTimestamp_{k_ary}_{n_fly}_{opaque_nbits}_{vc}_{payload_nbits}"

  if vc > 1:
    VcIdType = mk_bits( clog2( vc ) )

    return mk_bitstruct( new_name, {
      'src':       IdType,
      'dst':       DstType,
      'opaque':    OpqType,
      'vc_id':     VcIdType,
      'payload':   PayloadType,
      'timestamp': TimestampType,
    })
  else:
    return mk_bitstruct( new_name, {
      'src':       IdType,
      'dst':       DstType,
      'opaque':    OpqType,
      'payload':   PayloadType,
      'timestamp': TimestampType,
    })



"""
==========================================================================
OutputUnitCL.py
==========================================================================
Cycle level implementation of output unit.

Author : Yanghui Ou
  Date : July 2, 2019
"""
from pymtl3 import *


class OutputUnitCL( Component ):

  def construct( s, PacketType, QueueType = None ):

    # Interface
    s.get  = CallerIfcCL( Type=PacketType )
    s.send = CallerIfcCL( Type=PacketType )
    s.QueueType = QueueType

    # If queue type is assigned
    if s.QueueType != None:
      # Component
      # TODO: add type for QueueType when pymtl3 has that.
      s.queue = QueueType()

      @s.update
      def up_ou_get_enq():
        if s.get.rdy() and s.queue.enq.rdy():
          s.queue.enq( s.get() )

      @s.update
      def up_ou_deq_send():
        if s.queue.deq.rdy() and s.send.rdy():
          s.send( s.queue.deq() )

    # No ouput queue
    else:
      @s.update
      def up_ou_get_send():
        if s.get.rdy() and s.send.rdy():
          s.send( s.get() )

  def line_trace( s ):
    return "{}(){}".format( s.get, s.send )



"""
==========================================================================
 CreditIfc.py
==========================================================================
Credit based interfaces.

Author : Yanghui Ou
  Date : June 10, 2019
"""
from ocnlib.rtl import Counter
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL, enrdy_to_str
from pymtl3.stdlib.rtl.arbiters import RoundRobinArbiterEn
from pymtl3.stdlib.rtl.Encoder import Encoder
from pymtl3.stdlib.rtl.queues import BypassQueueRTL

#-------------------------------------------------------------------------
# RTL interfaces
#-------------------------------------------------------------------------

class CreditRecvIfcRTL( Interface ):

  def construct( s, MsgType, vc=1 ):
    assert vc > 1, "We only support multiple virtual channels!"

    s.en  = InPort ( Bits1   )
    s.msg = InPort ( MsgType )
    s.yum = [ OutPort( Bits1 ) for i in range( vc ) ]

    s.MsgType = MsgType
    s.vc    = vc

  def __str__( s ):
    try:
      trace_len = s.trace_len
    except AttributeError:
      s.trace_len = len( str(s.MsgType()) )
      trace_len = s.trace_len
    return "{}:{}".format(
      enrdy_to_str( s.msg, s.en, True, s.trace_len ),
      "".join( [ "$" if x else '.' for x in s.yum ] )
    )

class CreditSendIfcRTL( Interface ):

  def construct( s, MsgType, vc=1 ):
    assert vc > 1, "We only support multiple virtual channels!"

    s.en  = OutPort( Bits1   )
    s.msg = OutPort( MsgType )
    s.yum = [ InPort( Bits1 ) for _ in range( vc ) ]

    s.MsgType = MsgType
    s.vc    = vc

  def __str__( s ):
    try:
      trace_len = s.trace_len
    except AttributeError:
      s.trace_len = len( str(s.MsgType()) )
      trace_len = s.trace_len
    return "{}:{}".format(
      enrdy_to_str( s.msg, s.en, True, s.trace_len ),
      "".join( [ "$" if s.yum[i] else '.' for i in range(s.vc) ] )
    )

#-------------------------------------------------------------------------
# CL interfaces
#-------------------------------------------------------------------------

class CreditSendIfcCL( Interface ):

  def construct( s, MsgType, vc=1 ):
    assert vc > 1, "We only support multiple virtual channels!"

    s.send_msg    = CallerPort( MsgType )
    s.recv_credit = [ CalleePort() for _ in range( vc ) ]

    s.MsgType = MsgType
    s.vc    = vc

  def __str__( s ):
    return ""

class CreditRecvIfcCL( Interface ):

  def construct( s, MsgType, vc=1 ):
    assert vc > 1, "We only support multiple virtual channels!"

    s.recv_msg    = CalleePort( MsgType )
    s.send_credit = [ CallerPort() for _ in range( vc ) ]

    s.MsgType = MsgType
    s.vc    = vc

  def __str__( s ):
    return ""

#-------------------------------------------------------------------------
# CreditIfc adapters
#-------------------------------------------------------------------------

class RecvRTL2CreditSendRTL( Component ):

  def construct( s, MsgType, vc=2, credit_line=1 ):
    assert vc > 1

    # Interface

    s.recv = RecvIfcRTL( MsgType )
    s.send = CreditSendIfcRTL( MsgType, vc )

    s.MsgType = MsgType
    s.vc    = vc

    # Components

    CreditType = mk_bits( clog2(credit_line+1) )
    VcIDType   = mk_bits( clog2( vc ) if vc > 1 else 1 )

    # FIXME: use multiple buffers to avoid deadlock.
    s.buffer = BypassQueueRTL( MsgType, num_entries=1 )
    s.credit = [ Counter( CreditType, credit_line ) for _ in range( vc ) ]

    s.recv           //=  s.buffer.enq
    s.buffer.deq.ret //=  s.send.msg

    @s.update
    def up_credit_send():
      s.send.en = b1(0)
      s.buffer.deq.en = b1(0)
      if s.buffer.deq.rdy:
        for i in range( vc ):
          if VcIDType(i) == s.buffer.deq.ret.vc_id and s.credit[i].count > CreditType(0):
            s.send.en = b1(1)
            s.buffer.deq.en = b1(1)

    @s.update
    def up_counter_decr():
      for i in range( vc ):
        s.credit[i].decr = s.send.en & ( VcIDType(i) == s.send.msg.vc_id )

    for i in range( vc ):
      s.credit[i].incr       //= s.send.yum[i]
      s.credit[i].load       //= b1(0)
      s.credit[i].load_value //= CreditType(0)

  def line_trace( s ):
    return "{}({},{}){}".format(
      s.recv,
      s.buffer.line_trace(),
      ",".join( [ str(s.credit[i].count) for i in range(s.vc) ] ),
      s.send,
    )

class CreditRecvRTL2SendRTL( Component ):

  def construct( s, MsgType, vc=2, credit_line=1, QType=BypassQueueRTL ):
    assert vc > 1

    # Interface

    s.recv = CreditRecvIfcRTL( MsgType, vc )
    s.send = SendIfcRTL( MsgType )

    s.MsgType = MsgType
    s.vc    = vc

    # Components

    CreditType = mk_bits( clog2(credit_line+1) )
    ArbReqType = mk_bits( vc )
    VcIDType   = mk_bits( clog2( vc ) if vc > 1 else 1 )

    s.buffers = [ QType( MsgType, num_entries=credit_line )
                  for _ in range( vc ) ]
    s.arbiter = RoundRobinArbiterEn( nreqs=vc )
    s.encoder = Encoder( in_nbits=vc, out_nbits=clog2(vc) )

    for i in range( vc ):
      s.buffers[i].enq.msg //= s.recv.msg
      s.buffers[i].deq.rdy //= s.arbiter.reqs[i]
    s.arbiter.grants //= s.encoder.in_
    s.arbiter.en     //= s.send.en

    @s.update
    def up_enq():
      if s.recv.en:
        for i in range( vc ):
          s.buffers[i].enq.en = s.recv.msg.vc_id == VcIDType(i)
      else:
        for i in range( vc ):
          s.buffers[i].enq.en = b1(0)

    @s.update
    def up_deq_and_send():
      for i in range( vc ):
        s.buffers[i].deq.en = b1(0)

      s.send.msg = s.buffers[ s.encoder.out ].deq.ret
      if s.send.rdy & ( s.arbiter.grants > ArbReqType(0) ):
        s.send.en = b1(1)
        s.buffers[ s.encoder.out ].deq.en = b1(1)
      else:
        s.send.en = b1(0)

    @s.update
    def up_yummy():
      for i in range( vc ):
        s.recv.yum[i] = s.buffers[i].deq.en

  def line_trace( s ):
    return "{}(){}".format( s.recv, s.send )



"""
==========================================================================
measure_packets.py
==========================================================================
Packets for network measurement.

Author : Yanghui Ou
  Date : Sep 24, 2019
"""
from pymtl3 import *

#-------------------------------------------------------------------------
# mk_ring_pkt
#-------------------------------------------------------------------------

def mk_ring_pkt( nrouters=4, vc=2, payload_nbits=32 ):
  id_type = mk_bits( clog2( nrouters ) )
  v_type  = mk_bits( clog2( vc ) )
  p_type  = mk_bits( payload_nbits )
  new_name = f'RingMeasurePkt_{nrouters}_{vc}_{payload_nbits}'

  def str_func( s ):
    return f"{s.src}>{s.dst}:{s.vc_id}:{'M' if s.measure else " "}:{s.payload}"

  return mk_bitstruct( new_name, {
      'src':     id_type,
      'dst':     id_type,
      'vc_id':   v_type,
      'measure': Bits1,
      'payload': p_type,
    },
    namespace = { '__str__': str_func }
  )

#-------------------------------------------------------------------------
# mk_ring_pkt
#-------------------------------------------------------------------------

def mk_mesh_pkt( ncols=2, nrows=2, vc=1, payload_nbits=32 ):
  x_type = mk_bits( clog2( ncols ) )
  y_type = mk_bits( clog2( nrows ) )
  p_type = mk_bits( payload_nbits  )

  new_name = f'MeshMeasurePkt_{ncols}_{nrows}_{vc}_{payload_nbits}'

  if vc > 1:
    v_type = mk_bits( clog2( vc ) )

    def str_func( s ):
      return f"{s.src}>{s.dst}:{s.vc_id}:{'M' if s.measure else " "}:{s.payload}"

    return mk_bitstruct( new_name, {
        'src_x':   x_type,
        'src_y':   y_type,
        'dst_x':   x_type,
        'dst_y':   y_type,
        'vc_id':   v_type,
        'measure', Bits1,
        'payload', p_type,
      },
      namespace = { '__str__': str_func }
    )
  else:
    def str_func( s ):
      return f"{s.src}>{s.dst}:{'M' if s.measure else " "}:{s.payload}"

    return mk_bitstruct( new_name, {
        'src_x':   x_type,
        'src_y':   y_type,
        'dst_x':   x_type,
        'dst_y':   y_type,
        'measure', Bits1,
        'payload', p_type,
      },
      namespace = { '__str__': str_func }
    )



#=========================================================================
# ChannelRTL.py
#=========================================================================
# CL channel module for connecting routers to form network. This simple
# channel has latency insensitive send/recv interfaces.
#
# Author : Cheng Tan, Yanghui Ou
#   Date : Mar 16, 2019

from ocnlib.ifcs.PhysicalDimension import PhysicalDimension
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL
from pymtl3.stdlib.rtl.queues import NormalQueueRTL


class ChannelRTL( Component ):
  def construct(s, PacketType, QueueType=NormalQueueRTL, latency=0 ):

    # Constant
    s.dim = PhysicalDimension()
    s.QueueType   = QueueType
    s.latency     = latency
    s.num_entries = 2

    # Interface
    s.recv  = RecvIfcRTL( PacketType )
    s.send  = SendIfcRTL( PacketType )

    if s.QueueType != None and s.latency > 0:

      # Component

      s.queues = [ s.QueueType( PacketType, s.num_entries )
                   for _ in range( s.latency ) ]

      # Connections

      s.recv.rdy //= s.queues[0].enq.rdy

      @s.update
      def process():
        s.queues[0].enq.msg = s.recv.msg
        s.queues[0].enq.en  = s.recv.en and s.queues[0].enq.rdy
        for i in range(s.latency - 1):
          s.queues[i+1].enq.msg = s.queues[i].deq.ret
          s.queues[i+1].enq.en  = s.queues[i].deq.rdy and s.queues[i+1].enq.rdy
          s.queues[i].deq.en    = s.queues[i+1].enq.en

        s.send.msg  = s.queues[s.latency-1].deq.ret
        s.send.en   = s.send.rdy and s.queues[s.latency-1].deq.rdy
        s.queues[s.latency-1].deq.en   = s.send.en

    else:

      # If latency==0 simply bypass

      s.recv //= s.send

  def line_trace( s ):
    if s.QueueType != None and s.latency != 0:
      trace = '>'
      for i in range( s.latency ):
        trace += s.queues[i].line_trace() + '>'
      return f"{s.recv.msg}({trace}){s.send.msg}"
    else:
      return f"{s.recv}(0){s.send}"

  def elaborate_physical( s ):
    s.dim.w = 250



#=========================================================================
# TorusNetworkCL.py
#=========================================================================
# Cycle level Torus network implementation.
#
# Author : Yanghui Ou
#   Date : Mar 21, 2019

from channel.ChannelCL import ChannelCL
from directions import *
from pymtl3 import *
from TorusRouterCL import TorusRouterCL


class TorusNetworkCL( Component ):
  def construct( s, PacketType, PositionType,
                    ncols=4, nrows=4, chl_lat=0 ):

    # Constants

    s.ncols    = ncols
    s.nrows   = nrows
    s.num_routers   = ncols * nrows
    num_channels    = nrows * ncols * 4
    s.num_terminals = s.num_routers

    # Interface

    s.recv = [ NonBlockingCallee() for _ in range(s.num_terminals) ]
    s.send = [ NonBlockingCaller() for _ in range(s.num_terminals) ]

    # Components

    s.routers  = [ TorusRouterCL( PacketType, PositionType )
                    for i in range( s.num_routers ) ]

    s.channels = [ ChannelCL( PacketType, latency = chl_lat)
                    for _ in range( num_channels ) ]

    # Connect routers in Torus

    chl_id  = 0
    for i in range (s.num_routers):

      s_idx = (i-nrows+s.num_routers) % s.num_routers
      s.routers[i].send[SOUTH] //= s.channels[chl_id].recv
      s.channels[chl_id].send  //= s.routers[s_idx].recv[NORTH]
      chl_id += 1

      n_idx = (i+nrows+s.num_routers) % s.num_routers
      s.routers[i].send[NORTH] //= s.channels[chl_id].recv
      s.channels[chl_id].send  //= s.routers[n_idx].recv[SOUTH]
      chl_id += 1

      w_idx = i - ( i % ncols - (i-1) % ncols )
      s.routers[i].send[WEST] //= s.channels[chl_id].recv
      s.channels[chl_id].send //= s.routers[w_idx].recv[EAST]
      chl_id += 1

      e_idx = i + (i+1) % ncols - i % ncols
      s.routers[i].send[EAST] //= s.channels[chl_id].recv
      s.channels[chl_id].send //= s.routers[e_idx].recv[WEST]
      chl_id += 1

      # Connect the self port (with Network Interface)

      s.recv[i] //= s.routers[i].recv[SELF]
      s.send[i] //= s.routers[i].send[SELF]

    @s.update
    def up_pos():
      for y in range( nrows ):
        for x in range( ncols ):
          idx = y * ncols + x
          s.routers[idx].pos = PositionType( x, y )

  def line_trace( s ):
    trace = [ "" for _ in range( s.num_terminals ) ]
    for i in range( s.num_terminals ):
      trace[i] += s.routers[i].line_trace()
    return "|".join( trace )

  def elaborate_physical( s ):
    # Initialize dimension for sub-modules.
    BOUNDARY = 10

    for i, r in enumerate( s.routers ):
      r.dim.x = BOUNDARY + i % s.ncols * ( r.dim.w + s.channels[0].dim.w )
      r.dim.y = BOUNDARY + i / s.ncols * ( r.dim.h + s.channels[0].dim.w )

    s.dim.w = 2 * BOUNDARY + s.ncols * ( r.dim.w + s.channels[0].dim.w )
    s.dim.h = 2 * BOUNDARY + s.nrows  * ( r.dim.h + s.channels[0].dim.w )



"""
=========================================================================
DORXRouteUnitRTL.py
=========================================================================
A DOR route unit with get/give interface.

Author : Yanghui Ou, Cheng Tan
  Date : Mar 25, 2019
"""
from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL

from .directions import *


class DORXMeshRouteUnitRTL( Component ):

  def construct( s, PacketType, PositionType, num_outports ):

    # Interface

    s.get  = GetIfcRTL( PacketType )
    s.give = [ GiveIfcRTL(PacketType) for _ in range (num_outports) ]
    s.pos  = InPort( PositionType )

    # Componets

    s.out_dir  = Wire( mk_bits( clog2(num_outports) ) )
    s.give_ens = Wire( mk_bits( num_outports ) )

    # Connections

    for i in range( num_outports ):
      s.get.ret     //= s.give[i].ret
      s.give_ens[i] //= s.give[i].en

    # Routing logic
    @s.update
    def up_ru_routing():

      s.out_dir = Bits3(0)
      for i in range( num_outports ):
        s.give[i].rdy = Bits1(0)

      if s.get.rdy:
        if s.pos.pos_x == s.get.ret.dst_x and s.pos.pos_y == s.get.ret.dst_y:
          s.out_dir = SELF
        elif s.get.ret.dst_x < s.pos.pos_x:
          s.out_dir = WEST
        elif s.get.ret.dst_x > s.pos.pos_x:
          s.out_dir = EAST
        elif s.get.ret.dst_y > s.pos.pos_y:
          s.out_dir = NORTH
        else:
          s.out_dir = SOUTH
        s.give[ s.out_dir ].rdy = Bits1(1)

    @s.update
    def up_ru_give_en():
      s.get.en = s.give_ens > Bits5(0)

  # Line trace
  def line_trace( s ):
    out_str = "|".join([ str(x) for x in s.give ])
    return f"{s.get}({s.out_dir}){out_str}"



"""
==========================================================================
TorusRouterRTL.py
==========================================================================
Torus router RTL model.

Author : Yanghui, Ou, Cheng Tan
  Date : June 30, 2019
"""
from ocnlib.ifcs.CreditIfc import CreditRecvIfcRTL, CreditSendIfcRTL
from pymtl3 import *
from router.InputUnitCreditRTL import InputUnitCreditRTL
from router.OutputUnitCreditRTL import OutputUnitCreditRTL
from router.SwitchUnitRTL import SwitchUnitRTL

from .DORYTorusRouteUnitRTL import DORYTorusRouteUnitRTL


class TorusRouterRTL( Component ):

  def construct( s, PacketType, PositionType,
                    InputUnitType=InputUnitCreditRTL,
                    RouteUnitType=DORYTorusRouteUnitRTL,
                    SwitchUnitType=SwitchUnitRTL,
                    OutputUnitType=OutputUnitCreditRTL,
                    ncols=2, nrows=2, vc=2, credit_line=2,
  ):

    s.num_inports  = 5
    s.num_outports = 5
    s.vc = vc
    s.num_route_units = s.num_inports * s.vc

    # Interface

    s.pos  = InPort( PositionType )
    s.recv = [ CreditRecvIfcRTL( PacketType, s.vc ) for _ in range( s.num_inports  ) ]
    s.send = [ CreditSendIfcRTL( PacketType, s.vc ) for _ in range( s.num_outports ) ]

    # Components

    s.input_units  = [ InputUnitType( PacketType, vc=vc, credit_line=credit_line )
                      for _ in range( s.num_inports ) ]

    s.route_units  = [ RouteUnitType( PacketType, PositionType, ncols, nrows )
                      for _ in range( s.num_route_units ) ]

    s.switch_units = [ SwitchUnitType( PacketType, s.num_route_units )
                      for _ in range( s.num_outports ) ]

    s.output_units = [ OutputUnitType( PacketType )
                      for _ in range( s.num_outports ) ]

    # Connection

    for i in range( s.num_inports ):
      s.recv[i] //= s.input_units[i].recv
      for j in range( s.vc ):
        ru_idx = i * s.vc + j
        s.input_units[i].give[j] //= s.route_units[ru_idx].get
        s.pos                    //= s.route_units[ru_idx].pos

    for i in range( s.num_route_units ):
      for j in range( s.num_outports ):
        s.route_units[i].give[j] //= s.switch_units[j].get[i]

    for j in range( s.num_outports ):
      s.switch_units[j].give //= s.output_units[j].get
      s.output_units[j].send //= s.send[j]

  # Line trace

  def line_trace( s ):
    return "{}({}){}".format(
      "|".join( [ f"{str(x)}" for x in s.recv ] ),
      s.pos,
      "|".join( [ f"{str(x)}" for x in s.send ] )
    )
  def elaborate_physical( s ):
    s.dim.w = 50
    s.dim.h = 50



"""
==========================================================================
InputUnitCL_test.py
==========================================================================
Test cases for InputUnitCL.

Author: Yanghui Ou
  Date: May 16, 2019
"""
import hypothesis
import pytest
from hypothesis import strategies as st

from ocnlib.utils import run_sim
from pymtl3 import *
from pymtl3.datatypes import strategies as pst
from pymtl3.stdlib.cl.queues import BypassQueueCL, NormalQueueCL, PipeQueueCL
from pymtl3.stdlib.test.test_sinks import TestSinkCL
from pymtl3.stdlib.test.test_srcs import TestSrcCL
from router.InputUnitCL import InputUnitCL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, src_msgs, sink_msgs ):

    s.src  = TestSrcCL  ( MsgType, src_msgs  )
    s.sink = TestSinkCL ( MsgType, sink_msgs )
    s.dut  = InputUnitCL( MsgType )

    # Connections
    s.src.send //= s.dut.recv

    @s.update
    def up_give_en():
      if s.dut.give.rdy() and s.sink.recv.rdy():
        s.sink.recv( s.dut.give() )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{} >>> {} >>> {}".format(
      s.src.line_trace(),
      s.dut.line_trace(),
      s.sink.line_trace(),
    )

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

class InputUnitCL_Tests:

  @classmethod
  def setup_class( cls ):
    cls.TestHarness = TestHarness
    cls.qtypes      = [ NormalQueueCL, BypassQueueCL, PipeQueueCL ]

  def test_normal2_simple( s ):
    test_msgs = [ b16( 4 ), b16( 1 ), b16( 2 ), b16( 3 ) ]
    arrival_time = [ 2, 3, 4, 5 ]
    th = s.TestHarness( Bits16, test_msgs, test_msgs )
    th.set_param( "top.sink.construct", arrival_time=arrival_time )
    run_sim( th )

  def test_hypothesis( s ):
    @hypothesis.settings( deadline=None )
    @hypothesis.given(
      qsize     = st.integers(1, 16),
      dwid      = st.integers(1, 32),
      sink_init = st.integers(0, 20),
      qtype     = st.sampled_from( s.qtypes ),
      test_msgs = st.data(),
    )
    def actual_test( dwid, qsize, qtype, sink_init, test_msgs ):
      msgs = test_msgs.draw( st.lists( pst.bits(dwid), min_size=1, max_size=100 ) )
      th = s.TestHarness( mk_bits(dwid), msgs, msgs )
      th.set_param( "top.sink.construct", initial_delay=sink_init )
      th.set_param( "top.dut.construct", QueueType = qtype )
      th.set_param( "top.dut.queue.construct", num_entries=qsize )
      run_sim( th, max_cycles=5000 )
    actual_test()



"""
==========================================================================
InputUnitRTL_test.py
==========================================================================
Test cases for InputUnitRTL.

Author: Yanghui Ou
  Date: Mar 24, 2019
"""
import pytest

from pymtl3 import *
from pymtl3.stdlib.rtl.queues import (BypassQueueRTL, NormalQueueRTL,
                                      PipeQueueRTL)
from pymtl3.stdlib.test import TestVectorSimulator
from pymtl3.stdlib.test.test_sinks import TestSinkRTL
from pymtl3.stdlib.test.test_srcs import TestSrcRTL
from router.InputUnitRTL import InputUnitRTL

from .InputUnitCL_test import InputUnitCL_Tests as BaseTests

#-------------------------------------------------------------------------
# TestVectorSimulator test
#-------------------------------------------------------------------------

def run_tv_test( dut, test_vectors ):

  # Define input/output functions

  def tv_in( dut, tv ):
    dut.recv.en  = tv[0]
    dut.recv.msg = tv[2]
    dut.give.en  = tv[3]

  def tv_out( dut, tv ):
    if tv[1] != '?': assert dut.recv.rdy == tv[1]
    if tv[4] != '?': assert dut.give.rdy == tv[4]
    if tv[5] != '?': assert dut.give.ret == tv[5]

  # Run the test

  sim = TestVectorSimulator( dut, test_vectors, tv_in, tv_out )
  sim.run_test()

def test_pipe_Bits():

  run_tv_test( InputUnitRTL( Bits32 ), [
    #  enq.en  enq.rdy enq.msg   deq.en  deq.rdy deq.ret
    [  b1(1),  b1(1),  b32(123), b1(0),  b1(0),    '?'    ],
    [  b1(1),  b1(1),  b32(345), b1(0),  b1(1),  b32(123) ],
    [  b1(0),  b1(0),  b32(567), b1(0),  b1(1),  b32(123) ],
    [  b1(0),  b1(0),  b32(567), b1(1),  b1(1),  b32(123) ],
    [  b1(0),  b1(1),  b32(567), b1(1),  b1(1),  b32(345) ],
    [  b1(1),  b1(1),  b32(567), b1(0),  b1(0),    '?'    ],
    [  b1(1),  b1(1),  b32(0  ), b1(1),  b1(1),  b32(567) ],
    [  b1(1),  b1(1),  b32(1  ), b1(1),  b1(1),  b32(0  ) ],
    [  b1(1),  b1(1),  b32(2  ), b1(1),  b1(1),  b32(1  ) ],
    [  b1(0),  b1(1),  b32(2  ), b1(1),  b1(1),  b32(2  ) ],
] )

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, src_msgs, sink_msgs ):

    s.src  = TestSrcRTL   ( MsgType, src_msgs )
    s.dut  = InputUnitRTL ( MsgType )
    s.sink = TestSinkRTL  ( MsgType, sink_msgs )

    # Connections
    s.src.send     //= s.dut.recv
    s.dut.give.ret //= s.sink.recv.msg

    @s.update
    def up_give_en():
      if s.dut.give.rdy and s.sink.recv.rdy:
        s.dut.give.en  = b1(1)
        s.sink.recv.en = b1(1)
      else:
        s.dut.give.en  = b1(0)
        s.sink.recv.en = b1(0)

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{} >>> {} >>> {}".format(
      s.src.line_trace(),
      s.dut.line_trace(),
      s.sink.line_trace(),
    )

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------


class InputUnitRTL_Tests( BaseTests ):

  @classmethod
  def setup_class( cls ):
    cls.TestHarness = TestHarness
    cls.qtypes      = [ NormalQueueRTL, BypassQueueRTL, PipeQueueRTL ]



'''
==========================================================================
GrantHoldArbiter.py
==========================================================================
A round-robin arbiter with grant-hold circuit.

Author : Yanghui Ou
  Date : Jan 22, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.rtl.arbiters import RoundRobinArbiterEn


class GrantHoldArbiter( Component ):

  def construct( s, nreqs ):
    BitsN = mk_bits( nreqs )

    # Interface
    s.reqs   = InPort ( BitsN )
    s.hold   = InPort ( Bits1 )
    s.grants = OutPort( BitsN )

    # Components
    s.arb    = RoundRobinArbiterEn( nreqs )
    s.last_r = Wire( BitsN )

    # Logic
    s.arb.reqs //= lambda: BitsN(0) if s.hold else s.reqs
    s.arb.en   //= lambda: ~s.hold
    s.grants   //= lambda: s.arb.grants if ~s.hold else s.last_r

    @s.update_ff
    def up_last_r():
      s.last_r <<= s.grants
    
  def line_trace( s ):
    hold = 'h' if s.hold else ' '
    return f'{str(s.reqs.bin())[2:]}({hold}){str(s.grants.bin())[2:]}'



'''
==========================================================================
MeshNetworkMflitRTL_test.py
==========================================================================
Test cases for multi-flit mesh.

Author : Yanghui Ou
  Date : Feb 14, 2020
'''
import pytest
import hypothesis
from hypothesis import strategies as st
from pymtl3 import *
from pymtl3.stdlib.test import mk_test_case_table
from ocnlib.utils import to_bitstruct, run_sim
from ocnlib.packets import MflitPacket as Packet
from ocnlib.test.test_srcs import MflitPacketSourceRTL as TestSource
from ocnlib.test.test_sinks import MflitPacketSinkRTL as TestSink
from ..MeshNetworkMflitRTL import MeshNetworkMflitRTL

#-------------------------------------------------------------------------
# TestHeader
#-------------------------------------------------------------------------

@bitstruct
class TestHeader:
  opaque: Bits8
  src_x : Bits4
  src_y : Bits4
  dst_x : Bits4
  dst_y : Bits4
  plen  : Bits8

#-------------------------------------------------------------------------
# TestPosition
#-------------------------------------------------------------------------

@bitstruct
class TestPosition:
  pos_x : Bits4
  pos_y : Bits4

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, Header, Position, ncols, nrows, src_pkts, sink_pkts ):
    nterminals = ncols * nrows

    s.src  = [ TestSource( Header, src_pkts[i] ) for i in range( nterminals ) ]
    s.dut  = MeshNetworkMflitRTL( Header, Position, ncols, nrows )
    s.sink = [ TestSink( Header, sink_pkts[i] ) for i in range( nterminals ) ]

    for i in range( nterminals ):
      s.src[i].send //= s.dut.recv[i]
      s.dut.send[i] //= s.sink[i].recv

  def done( s ):
    src_done = True
    sink_done = True
    for m in s.src:
      src_done &= m.done()
    for m in s.sink:
      sink_done &= m.done()
    return src_done and sink_done

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# mk_pkt
#-------------------------------------------------------------------------

def mk_pkt( src_x, src_y, dst_x, dst_y, payload=[], opaque=0 ):
  plen        = len( payload )
  header      = TestHeader( opaque, src_x, src_y, dst_x, dst_y, plen )
  header_bits = to_bits( header )
  flits       = [ header_bits ] + payload
  return Packet( TestHeader, flits )

#-------------------------------------------------------------------------
# arrange_src_sink_pkts
#-------------------------------------------------------------------------

def arrange_src_sink_pkts( Header, ncols, nrows, pkt_lst ):
  nterminals = ncols * nrows
  src_pkts  = [ [] for _ in range( nterminals ) ]
  sink_pkts = [ [] for _ in range( nterminals ) ]

  for pkt in pkt_lst:
    header = to_bitstruct( pkt.flits[0], Header )
    src_id = header.src_x.uint() + header.src_y.uint() * ncols
    dst_id = header.dst_x.uint() + header.dst_y.uint() * ncols
    src_pkts [ src_id ].append( pkt )
    sink_pkts[ dst_id ].append( pkt )

  return src_pkts, sink_pkts

#-------------------------------------------------------------------------
# sanity check
#-------------------------------------------------------------------------

def test_sanity_check():
  net = MeshNetworkMflitRTL( TestHeader, TestPosition, ncols=2, nrows=2 )
  net.elaborate()
  net.apply( SimulationPass() )
  net.sim_reset()
  net.tick()

#-------------------------------------------------------------------------
# test case: basic
#-------------------------------------------------------------------------

def basic_pkts( ncols, nrows ):
  return [
    #       src_x    y        dst_x    y        payload
    mk_pkt(     0,   0,       ncols-1, nrows-1, [ 0x8badf00d             ] ),
    mk_pkt( ncols-1, nrows-1, 0,       0,       [ 0x8badbabe, 0xdeadbeef ] ),
  ]

#-------------------------------------------------------------------------
# test case: basic
#-------------------------------------------------------------------------

def neighbor_pkts( ncols, nrows ):
  pkts = []
  nterminals = ncols * nrows
  for i in range( nterminals ):
    src_x = i %  ncols
    src_y = i // ncols
    dst_x = ( i+1 ) %  ncols
    dst_y = ( ( i+1 ) % nterminals ) // ncols
    payload = [ x for x in range( i % 10 ) ]
    pkts.append( mk_pkt( src_x, src_y, dst_x, dst_y, payload ) )

  return pkts

#-------------------------------------------------------------------------
# test case table
#-------------------------------------------------------------------------

test_case_table = mk_test_case_table([
  (                'msg_func        ncols  nrows' ),
  [ 'basic',        basic_pkts,     2,     2      ],
  [ 'basic4x4',     basic_pkts,     4,     4      ],
  [ 'neighbor2x2',  neighbor_pkts,  2,     2      ],
  [ 'neighbor2x3',  neighbor_pkts,  2,     3      ],
  [ 'neighbor3x3',  neighbor_pkts,  3,     3      ],
  [ 'neighbor4x4',  neighbor_pkts,  4,     4      ],
])

#-------------------------------------------------------------------------
# test driver
#-------------------------------------------------------------------------

@pytest.mark.parametrize( **test_case_table )
def test_mflit_mesh( test_params, test_verilog ):
  ncols = test_params.ncols
  nrows = test_params.nrows
  pkts  = test_params.msg_func( ncols, nrows )
  trans_backend = 'verilog' if test_verilog else ''

  src_pkts, dst_pkts = arrange_src_sink_pkts( TestHeader, ncols, nrows, pkts )

  th = TestHarness( TestHeader, TestPosition, ncols, nrows,
                    src_pkts, dst_pkts )
  run_sim( th, translation=trans_backend )

#-------------------------------------------------------------------------
# packet strategy
#-------------------------------------------------------------------------

hex_words = [ 0x8badf00d, 0xdeadbeef, 0xfaceb00c, 0xdeadc0de ]

@st.composite
def pkt_strat( draw, ncols, nrows, max_plen=15 ):
  payload = draw( st.lists( st.sampled_from( hex_words ), min_size=0, max_size=max_plen ) )
  src_x   = draw( st.integers(0, ncols-1) )
  src_y   = draw( st.integers(0, nrows-1) )
  dst_x   = draw( st.integers(0, ncols-1) )
  dst_y   = draw( st.integers(0, nrows-1) )
  return mk_pkt( src_x, src_y, dst_x, dst_y, payload )

#-------------------------------------------------------------------------
# pyh2 test
#-------------------------------------------------------------------------

@hypothesis.settings( deadline=None, max_examples=50 )
@hypothesis.given(
  ncols = st.integers(2, 4),
  nrows = st.integers(2, 4),
  pkts  = st.data(),
)
def test_pyh2( ncols, nrows, pkts, test_verilog ):
  pkts = pkts.draw( st.lists( pkt_strat( ncols, nrows ), min_size=1, max_size=100 ) )
  src_pkts, dst_pkts = arrange_src_sink_pkts( TestHeader, ncols, nrows, pkts )
  trans_backend = 'verilog' if test_verilog else ''
  th = TestHarness( TestHeader, TestPosition, ncols, nrows,
                    src_pkts, dst_pkts )
  run_sim( th, translation=trans_backend )



#=========================================================================
# test_helpers.py
#=========================================================================
# Helper functions for meshnet tests.
#
# Author : Yanghui Ou
#   Date : Apr 30, 2019

from meshnet.directions import *


def dor_routing( src_x, src_y, dst_x, dst_y,
                 pos_x, pos_y, dimension='x' ):
  tsrc  = 0
  tsink = 0

  # Compute src port
  if src_x == pos_x and src_y == pos_y:
    tsrc = SELF

  elif dimension.lower() == 'y':
    if   src_x == pos_x: tsrc = SOUTH if src_y < pos_y else NORTH
    elif src_x < pos_x:  tsrc = WEST
    else:                tsrc = EAST
      
  elif dimension.lower() == 'x':
    if   src_y == pos_y: tsrc = WEST if src_x < pos_x else EAST
    elif src_y < pos_y:  tsrc = SOUTH
    else:                tsrc = NORTH

  else:
    raise AssertionError( "dimension must be either x or y!" )

  # Compute dst port
  if dst_x == pos_x and dst_y == pos_y:
    tsink = SELF

  elif dimension.lower() == 'y':
    tsink = ( 
      NORTH if dst_y > pos_y else
      SOUTH if dst_y < pos_y else
      EAST  if dst_x > pos_x else
      WEST
    )

  elif dimension.lower() == 'x':
    tsink = ( 
      EAST  if dst_x > pos_x else
      WEST  if dst_x < pos_x else
      NORTH if dst_y > pos_y else
      SOUTH 
    )

  else:
    raise AssertionError( "dimension must be either x or y!" )

  return tsrc, tsink



#=========================================================================
# positions.py
#=========================================================================
# Positions definition.
#
# Author : Yanghui Ou
#   Date : May 20, 2019

from pymtl3 import *


def mk_mesh_pos( ncols, nrows ):
  assert ncols > 0 and nrows > 0

  XType = mk_bits(clog2( ncols  )) if ncols  != 1 else Bits1
  YType = mk_bits(clog2( nrows )) if nrows != 1 else Bits1

  return mk_bitstruct( f"MeshPosition_{ncols}x{nrows}", {
    'pos_x': XType,
    'pos_y': YType
  })

def mk_ring_pos( nrouters ):
  assert nrouters > 1
  return mk_bits( clog2( nrouters ) )

#=========================================================================
# Butterfly position
#=========================================================================

def mk_bfly_pos( k_ary=2, n_fly=2 ):
  assert n_fly > 0 and k_ary > 0

  row = k_ary ** ( n_fly - 1 )
  NflyType = mk_bits( clog2( n_fly ) ) if n_fly != 1 else Bits1
  RrowType = mk_bits( clog2( k_ary ** ( n_fly - 1 ) ) ) if row != 1 else Bits1

  return mk_bitstruct( f"BflyPosition_{k_ary}_{n_fly}", {
    'row':   RrowType,
    'stage': NflyType,
  })



'''
==========================================================================
connects.py
==========================================================================
Collection of customized connects.

Author : Yanghui Ou
  Date : Jan 29, 2020
'''
from pymtl3 import *
from pymtl3.datatypes.bitstructs import (
  is_bitstruct_class, 
  _FIELDS as bitstruct_fields,
)

#------------------------------------------------------------------------- 
# bitstruct_to_slice_h
#------------------------------------------------------------------------- 
# Recursive helper function

def _bitstruct_to_slices_h( cls, slices ): 
  if issubclass( cls, Bits ):
    start = 0 if not slices else slices[-1].stop
    stop  = start + cls.nbits
    slices.append( slice( start, stop ) )

  else:
    assert is_bitstruct_class( cls )
    fields_dict = getattr( cls, bitstruct_fields )
    fields_lst  = list( fields_dict.items() )
    # Use reverse instead of [::-1] because it is faster
    fields_lst.reverse()
    for _, ftype in fields_lst:
      _bitstruct_to_slices_h( ftype, slices )

#------------------------------------------------------------------------- 
# bitstruct_to_slice
#------------------------------------------------------------------------- 
# Converts a bitstruct to a list of slices
# TODO: import this from commons.py

def bitstruct_to_slices( cls ):
  slices = []
  _bitstruct_to_slices_h( cls, slices )
  return slices

#------------------------------------------------------------------------- 
# _connect_bitstruct_h
#------------------------------------------------------------------------- 

def _connect_bitstruct_h( field, bits_signal, slices ): 
  field_type = field._dsl.Type

  if issubclass( field_type, Bits ):
    start     = 0 if not slices else slices[-1].stop
    stop      = start + field_type.nbits
    cur_slice = slice( start, stop )
    slices.append( cur_slice )
    connect( field, bits_signal[ cur_slice ] )

  else:
    assert is_bitstruct_class( field_type )
    fields_dict = getattr( field_type, bitstruct_fields )
    fields_lst  = list( fields_dict.items() )
    # Use reverse instead of [::-1] because it is faster
    fields_lst.reverse()
    for fname, _ in fields_lst:
      subfield = getattr( field, fname )
      _connect_bitstruct_h( subfield, bits_signal, slices )

#------------------------------------------------------------------------- 
# connect_bitstruct
#------------------------------------------------------------------------- 

def connect_bitstruct( signal1, signal2 ):
  type1 = signal1._dsl.Type
  type2 = signal2._dsl.Type

  if is_bitstruct_class( type1 ):
    assert issubclass( type2, Bits )
    bits_signal, bitstruct_signal = signal2, signal1
  elif is_bitstruct_class( type2 ):
    assert issubclass( type1, Bits )
    bits_signal, bitstruct_signal = signal1, signal2
  else:
    assert False, "Can only connect a bitstruct wire to bits wire"

  # Connect field to corresponding slice
  _connect_bitstruct_h( bitstruct_signal, bits_signal, [] )





"""
=========================================================================
MeshNetworkRTL_test.py
=========================================================================
Test for NetworkRTL

Author : Yanghui Ou, Cheng Tan
  Date : Mar 20, 2019
"""
from meshnet.DORXMeshRouteUnitRTL import DORXMeshRouteUnitRTL
from meshnet.DORYMeshRouteUnitRTL import DORYMeshRouteUnitRTL
from meshnet.MeshNetworkRTL import MeshNetworkRTL
from ocnlib.ifcs.packets import mk_mesh_pkt
from ocnlib.ifcs.positions import mk_mesh_pos
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkRTL
from pymtl3 import *
from pymtl3.stdlib.rtl.queues import NormalQueueRTL
from pymtl3.stdlib.test import TestVectorSimulator
from pymtl3.stdlib.test.test_srcs import TestSrcRTL
from router.InputUnitRTL import InputUnitRTL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, ncols, nrows, src_msgs, sink_msgs,
                 src_initial, src_interval, sink_initial, sink_interval,
                 arrival_time=None ):

    MeshPos = mk_mesh_pos( ncols, nrows )
    s.num_routers = ncols * nrows
    s.dut = MeshNetworkRTL( MsgType, MeshPos, ncols, nrows, 0)
    match_func = lambda a, b : a.payload == b.payload

    s.srcs  = [ TestSrcRTL   ( MsgType, src_msgs[i],  src_initial,  src_interval  )
                for i in range ( s.dut.num_routers ) ]
    if arrival_time != None:
      s.sinks = [ TestNetSinkRTL( MsgType, sink_msgs[i], sink_initial,
                  sink_interval, arrival_time[i], match_func=match_func)
                  for i in range ( s.dut.num_routers ) ]
    else:
      s.sinks = [ TestNetSinkRTL  ( MsgType, sink_msgs[i], sink_initial,
                  sink_interval, match_func=match_func)
                  for i in range ( s.dut.num_routers ) ]

    # Connections

    for i in range ( s.dut.num_routers ):
      s.srcs[i].send //= s.dut.recv[i]
      s.dut.send[i]  //= s.sinks[i].recv

  def done( s ):
    srcs_done = 1
    sinks_done = 1
    for i in range( s.num_routers ):
      if s.srcs[i].done() == 0:
        srcs_done = 0
    for i in range( s.num_routers ):
      if s.sinks[i].done() == 0:
        sinks_done = 0
    return srcs_done and sinks_done
  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# Test cases (specific for 4x4 mesh)
#-------------------------------------------------------------------------

def test_srcsink_mesh4x4():

  #           src, dst, payload
  test_msgs = [ (0, 15, 101), (1, 14, 102), (2, 13, 103), (3, 12, 104),
                (4, 11, 105), (5, 10, 106), (6,  9, 107), (7,  8, 108),
                (8,  7, 109), (9,  6, 110), (10, 5, 111), (11, 4, 112),
                (12, 3, 113), (13, 2, 114), (14, 1, 115), (15, 0, 116) ]

  src_packets  =  [ [],[],[],[],
                    [],[],[],[],
                    [],[],[],[],
                    [],[],[],[] ]

  sink_packets =  [ [],[],[],[],
                    [],[],[],[],
                    [],[],[],[],
                    [],[],[],[] ]

  # note that need to yield one/two cycle for reset
  arrival_pipes = [[8], [6], [6], [8],
                   [6], [4], [4], [6],
                   [6], [4], [4], [6],
                   [8], [6], [6], [8]]

  ncols = 4
  nrows  = 4

  PacketType = mk_mesh_pkt( ncols, nrows )
  for (src, dst, payload) in test_msgs:
    pkt = PacketType( src%ncols, src//ncols, dst%ncols, dst//ncols, 1, payload )
    src_packets [src].append( pkt )
    sink_packets[dst].append( pkt )

  for i,x in enumerate(src_packets):
    print("src", i, [str(y) for y in x] )
  for i,x in enumerate(sink_packets):
    print("sink", i, [str(y) for y in x] )
  th = TestHarness( PacketType, ncols, nrows, src_packets, sink_packets,
                    0, 0, 0, 0, arrival_pipes )
  run_sim( th )



"""
==========================================================================
DTRBfRouteUnitRTL.py
==========================================================================
RTL implementation of a route unit for butterfly network. It uses
destination tag routing.

Author : Yanghui Ou, Cheng Tan
  Date : April 6, 2019
"""
from copy import deepcopy

from pymtl3 import *
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL


class DTRBflyRouteUnitRTL( Component ):

  def construct( s, PacketType, PositionType, num_outports, n_fly = 3 ):

    # Constants

    s.num_outports = num_outports
    k_ary    = num_outports
    OutType  = mk_bits( clog2( s.num_outports ) )
    rows     = k_ary ** ( n_fly - 1 )
    EnType  = mk_bits( s.num_outports )
    DstType = mk_bits( clog2(k_ary) * n_fly )
    RowWidth = clog2( k_ary )
    END = n_fly * RowWidth
    BEGIN = END - RowWidth

    # Interface

    s.get  = GetIfcRTL( PacketType )
    s.give = [ GiveIfcRTL(PacketType ) for _ in range ( s.num_outports ) ]
    s.pos  = InPort( PositionType )

    # Componets

    s.out_dir  = Wire( OutType )
    s.give_rdy = [ Wire( Bits1 ) for _ in range( s.num_outports ) ]
    s.give_ens = Wire( mk_bits( s.num_outports ) )

    # Connections

    for i in range( s.num_outports ):
      s.give_ens[i] //= s.give[i].en
      s.give_rdy[i] //= s.give[i].rdy

    # Routing logic

    @s.update
    def up_ru_routing():
      for i in range( s.num_outports ):
        s.give_rdy[i] = b1(0)

      if s.get.rdy:
        s.out_dir = s.get.ret.dst[ BEGIN : END]
        s.give_rdy[ s.out_dir ] = b1(1)

    @s.update
    def up_ru_get_en():
      s.get.en = s.give_ens > EnType(0)
      for i in range( s.num_outports ):
        s.give[i].ret = deepcopy( s.get.ret )
      if s.get.rdy:
        s.give[ s.out_dir ].ret.dst = DstType(s.get.ret.dst << RowWidth)

  # Line trace

  def line_trace( s ):
    out_str = [ "" for _ in range( s.num_outports ) ]
    for i in range (s.num_outports):
      out_str[i] = "{}".format( s.give[i] )

    return "{}({}){}".format( s.get, s.out_dir, "|".join( out_str ) )



'''
==========================================================================
DeserializerRTL.py
==========================================================================
A Generic deserializer.

Author : Yanghui Ou
  Date : Feb 26, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL

from .Counter import Counter

class DeserializerRTL( Component ):

  #-----------------------------------------------------------------------
  # construct
  #-----------------------------------------------------------------------

  def construct( s, in_nbits, max_nblocks ):

    # Local parameter

    InType    = mk_bits( in_nbits )
    OutType   = mk_bits( in_nbits * max_nblocks )
    CountType = mk_bits( clog2( max_nblocks +1 ) )
    SelType   = mk_bits( clog2( max_nblocks ) )

    s.STATE_IDLE = b1(0)
    s.STATE_RECV = b1(1)

    # Interface

    s.recv = RecvIfcRTL( InType    )
    s.len  = InPort    ( CountType )
    s.send = SendIfcRTL( OutType   )

    # Component

    s.state      = Wire( Bits1 )
    s.state_next = Wire( Bits1 )
    s.idx        = Wire( CountType )
    s.len_r      = Wire( CountType )
    s.out_r      = [ Wire( InType ) for _ in range( max_nblocks ) ]
    s.counter    = Counter( CountType )( decr=b1(0) )

    s.idx //= s.counter.count

    # Input register

    @s.update_ff
    def up_len_r():
      if ( s.state == s.STATE_IDLE ) & s.recv.en | \
         ( s.state == s.STATE_RECV ) & ( s.idx == s.len_r ) & s.send.en:
        s.len_r <<= s.len if s.len > CountType(0) else CountType(1)
      else:
        s.len_r <<= CountType(0) if s.state_next == s.STATE_IDLE else s.len_r

    # Reg write logic

    @s.update_ff
    def up_out_r():
      if s.reset:
        for i in range( max_nblocks ):
          s.out_r[i] <<= InType(0)

      elif ( s.state == s.STATE_RECV ) & ( s.idx == s.len_r ) & s.send.en:
        if s.recv.en:
          s.out_r[0] <<= s.recv.msg
        else:
          s.out_r[0] <<= InType(0)
        for i in range(1, max_nblocks):
          s.out_r[i] <<= InType(0)

      elif s.recv.en:
        s.out_r[ SelType( s.idx ) ] <<= s.recv.msg

    for i in range( max_nblocks ):
      s.send.msg[i*in_nbits:(i+1)*in_nbits] //= s.out_r[i]

    # Counter logic

    s.counter.incr //= s.recv.en

    # Recv/Send logic

    s.send.en  //= lambda: ( s.state == s.STATE_RECV ) & ( s.idx ==  s.len_r ) & s.send.rdy

    @s.update
    def up_recv_rdy():
      if s.state == s.STATE_IDLE:
        s.recv.rdy = b1(1)

      else: # STATE_RECV
        if s.idx < s.len_r:
          s.recv.rdy = b1(1)
        elif s.send.en:
          s.recv.rdy = b1(1)

        else:
          s.recv.rdy = b1(0)

    # State transition logic

    @s.update_ff
    def up_state():
      if s.reset:
        s.state <<= s.STATE_IDLE
      else:
        s.state <<= s.state_next

    @s.update
    def up_state_next():
      if s.state == s.STATE_IDLE:
        if ( s.len > CountType(0) ) & s.recv.en:
          s.state_next         = s.STATE_RECV
          s.counter.load       = b1(1)
          s.counter.load_value = CountType(1)

        else:
          s.state_next   = s.STATE_IDLE
          s.counter.load = b1(0)

      else: # STATE_RECV
        if ( s.idx == s.len_r ) & s.send.en:
          if s.recv.en:
            s.state_next         = s.STATE_RECV
            s.counter.load       = b1(1)
            s.counter.load_value = CountType(1)
          else:
            s.state_next         = s.STATE_IDLE
            s.counter.load       = b1(1)
            s.counter.load_value = CountType(0)

        else:
          s.state_next         = s.STATE_RECV
          s.counter.load       = b1(0)
          s.counter.load_value = CountType(0)

  #-----------------------------------------------------------------------
  # line_trace
  #-----------------------------------------------------------------------

  def line_trace( s ):
    state = 'I' if s.state == s.STATE_IDLE else \
            'R' if s.state == s.STATE_RECV else \
            '?'
    return f'{s.recv}({state}:{s.idx}<{s.len_r}){s.send}'



"""
=========================================================================
BfNetworkRTL.py
=========================================================================
RTL implementation of a butterfly network.

Author : Cheng Tan
  Date : April 6, 2019
"""
from channel.ChannelRTL import ChannelRTL
from ocnlib.ifcs.PhysicalDimension import PhysicalDimension
from pymtl3 import *
from pymtl3.stdlib.ifcs.SendRecvIfc import *

from .BflyRouterRTL import BflyRouterRTL


class BflyNetworkRTL( Component ):

  def construct( s, PacketType, PositionType, k_ary, n_fly, chl_lat=0 ):

    # Constants

    s.dim           = PhysicalDimension()
    s.k_ary         = k_ary
    s.n_fly         = n_fly
    s.r_rows        = k_ary ** ( n_fly - 1 )
    s.num_routers   = n_fly * ( s.r_rows )
    s.num_terminals = k_ary ** n_fly
    num_channels    = ( n_fly - 1 ) * ( s.r_rows ) * k_ary

    # Interface

    s.recv = [ RecvIfcRTL(PacketType) for _ in range(s.num_terminals)]
    s.send = [ SendIfcRTL(PacketType) for _ in range(s.num_terminals)]

    # Components

    s.routers  = [ BflyRouterRTL( PacketType, PositionType, k_ary, n_fly )
                   for i in range( s.num_routers ) ]

    s.channels = [ ChannelRTL( PacketType, latency = chl_lat )
                 for _ in range( num_channels) ]

    s.pos      = [ PositionType( r%s.r_rows, r//s.r_rows )
                 for r in range( s.num_routers )]

    # Connect routers together in Butterfly

    for r in range( s.num_routers ):
      s.routers[r].pos.row   //= s.pos[r].row
      s.routers[r].pos.stage //= s.pos[r].stage

    chl_id = 0
    terminal_id_recv = 0
    terminal_id_send = 0
    group_size = s.r_rows

    for f in range( n_fly - 1 ):
      num_group  = s.r_rows // group_size
      for g in range( num_group ):
        for gs in range( group_size ):
          for k in range( k_ary ):
            index = g * group_size + gs
            base  = g * group_size
            interval = group_size // k_ary * k

            router_left   = f*s.r_rows+index
            router_right  = (f+1)*s.r_rows+base+(gs+interval)%group_size
            group_left    = router_left  % s.r_rows
            group_right   = router_right % s.r_rows
            group_left_x  = group_left   % k_ary
            group_right_x = group_right  % k_ary
            group_left_y  = group_left  // k_ary
            group_right_y = group_right // k_ary

            rtr_idx   = f * s.r_rows + index
            ifc_tdx   = ( k + gs // (group_size//k_ary) ) % k_ary
            rtr_idx_n = (f+1)*s.r_rows+base+(gs+interval)%group_size

            s.routers[rtr_idx].send[ifc_tdx] //= s.channels[chl_id].recv
            s.channels[chl_id].send          //= s.routers[rtr_idx_n].recv[k]

            chl_id += 1
      group_size = group_size // k_ary

    # Connect the router ports with Network Interfaces

    for i in range( s.num_routers ):
      if i < s.r_rows:
        for j in range( k_ary ):
          s.recv[terminal_id_recv] //= s.routers[i].recv[j]
          terminal_id_recv += 1

      if i >= s.num_routers - s.r_rows:
        for j in range( k_ary ):
          s.send[terminal_id_send] //= s.routers[i].send[j]
          terminal_id_send += 1

  def line_trace( s ):
      send_lst = []
      recv_lst = []
      for r in s.routers:
        has_recv = any([ r.recv[i].en for i in range(s.k_ary) ])
        has_send = any([ r.send[i].en for i in range(s.k_ary) ])
        if has_send:
          send_lst.append( f'{r.pos}' )
        if has_recv:
          recv_lst.append( f'{r.pos}')
      send_str = ','.join( send_lst )
      recv_str = ','.join( recv_lst )
      return f' {send_str:2} -> {recv_str:2}'

  def elaborate_physical( s ):

    for r in s.routers:
      r.elaborate_physical()

    BOUNDARY = 500
    INTERVAL_MINOR = 20
    INTER_ROUTER = 1000

    router_length = s.routers[0].dim.w
    router_height = s.routers[0].dim.h

    print( 'setPlaceMode -hardFence  true' )
    print( 'setCTSMode   -honorFence true' )
    print( 'setOptMode   -honorFence true' )
    MAX_X = 0
    MAX_Y = 0

    for row in range( s.r_rows ):
      for f in range( s.n_fly ):
        r = s.routers[ row + f * s.r_rows ]

        r.dim.x = BOUNDARY + row%s.k_ary * (s.n_fly*router_length + \
                  INTER_ROUTER + (s.n_fly-1)*INTERVAL_MINOR) + \
                  f * (router_length + INTERVAL_MINOR )
        r.dim.y = BOUNDARY + (s.r_rows//s.k_ary - 1 - row//s.k_ary) * \
                  (router_height + INTER_ROUTER)
        print( 'createFence routers___{}  {}  {}  {}  {}'.\
              format( row+f*s.r_rows, r.dim.x, r.dim.y,\
                      r.dim.x+router_length, r.dim.y+router_height ))

        if r.dim.x+router_length > MAX_X:
          MAX_X = r.dim.x + router_length

        if r.dim.y+router_height > MAX_Y:
          MAX_Y = r.dim.y + router_height

    print( 'size: {},{}'.format( MAX_X + BOUNDARY, MAX_Y + BOUNDARY ) )

  def elaborate_logical( s ):
    link_length = s.channels[0].dim.w

    for i, r in enumerate( s.routers ):
      r.dim.x = i // s.r_rows * ( r.dim.w + link_length )
      r.dim.y = i %  s.r_rows * ( r.dim.h + link_length )

    s.dim.w = s.r_rows * ( r.dim.w + link_length )
    s.dim.h = s.r_rows * ( r.dim.h + link_length )



#=========================================================================
# DORYMeshRouteUnitCL.py
#=========================================================================
# Cycle level implementation of a Y-DOR route unit.
#
# Author : Yanghui Ou
#   Date : May 21, 2019

from directions import *
from pymtl3 import *
from pymtl3.stdlib.cl.queues import BypassQueueCL


class DORYMeshRouteUnitCL( Component ):

  def construct( s, PacketType, PositionType ):

    # Local parameters

    s.num_outports = 5

    # Interface

    s.get  = CallerIfcCL( Type=PacketType )
    s.give = [ CalleeIfcCL( Type=PacketType ) for _ in range( s.num_outports ) ]
    s.pos  = InPort( PositionType )

    # Components

    s.rdy_lst = [ False for _ in range( s.num_outports ) ]
    s.pkt     = None

    @s.update
    def ru_up_route():
      if s.pkt is None and s.get.rdy():
        s.pkt = s.get()
      s.rdy_lst = [ False for _ in range( s.num_outports ) ]
      if s.pkt is not None:
        if s.pkt.dst_x == s.pos.pos_x and s.pkt.dst_y == s.pos.pos_y:
          s.rdy_lst[SELF] = True
        elif s.pkt.dst_y > s.pos.pos_y:
          s.rdy_lst[NORTH] = True
        elif s.pkt.dst_y < s.pos.pos_y:
          s.rdy_lst[SOUTH] = True
        elif s.pkt.dst_x < s.pos.pos_x:
          s.rdy_lst[WEST] = True
        else:
          s.rdy_lst[EAST] = True

    # Assign method and ready

    for i in range( s.num_outports ):
      def gen_give_rdy( s, port_id ):
        def give_rdy():
          if s.pkt is not None:
            return s.rdy_lst[port_id]
          else:
            return False
        return give_rdy

      s.give[i].rdy.method = gen_give_rdy( s, i )
      s.give[i].method.method = s.give_method

    for i in range( s.num_outports ):
      s.add_constraints(
        M( s.get ) < U( ru_up_route ) < M( s.give[i] ),
      )


  def give_method( s ):
    assert s.pkt is not None
    ret = s.pkt
    s.pkt = None
    return ret

  def line_trace( s ):
    out_trace = "|".join([ str(s.give[i]) for i in range(s.num_outports) ])
    return f"{s.get}(){out_trace}"



'''
==========================================================================
SwitchUnitGrantHoldRTL.py
==========================================================================
Switch unit that supports multi-flit (single-phit flit) packet.

R.I.P. Kobe.

Author : Yanghui Ou
  Date : Jan 26, 2020
'''
from pymtl3 import *
from pymtl3.stdlib.rtl import Mux
from pymtl3.stdlib.rtl.Encoder import Encoder
from pymtl3.stdlib.ifcs import GetIfcRTL, GiveIfcRTL, SendIfcRTL
from ocnlib.rtl import Counter, GrantHoldArbiter

class SwitchUnitGrantHoldRTL( Component ):

  def construct( s, Type, num_inports=5 ):

    # Local parameters
    s.num_inports = num_inports
    s.Type        = Type
    s.sel_width   = clog2( num_inports )

    GrantType     = mk_bits( num_inports )
    SelType       = mk_bits( s.sel_width )

    # Interface
    s.get  = [ GetIfcRTL( s.Type ) for _ in range( num_inports )  ]
    s.hold = [ InPort( Bits1 ) for _ in range( num_inports ) ]
    s.give = GiveIfcRTL( s.Type )

    # Components
    s.granted_get_rdy = Wire( Bits1 )
    s.any_hold        = Wire( Bits1 )

    s.arbiter = GrantHoldArbiter( nreqs=num_inports )( hold = s.any_hold )
    s.mux     = Mux( s.Type, num_inports )( out = s.give.ret )
    s.encoder = Encoder( num_inports, s.sel_width )(
      in_ = s.arbiter.grants,
      out = s.mux.sel,
    )

    # Combinational Logic
    @s.update
    def up_any_hold():
      s.any_hold = b1(0)
      for i in range( num_inports ):
        if s.hold[i]:
          s.any_hold = b1(1)

    @s.update
    def up_granted_get_rdy():
      s.granted_get_rdy = b1(0)
      for i in range( num_inports ):
        if s.arbiter.grants[i]:
          s.granted_get_rdy = s.get[i].rdy

    for i in range( num_inports ):
      s.get[i].rdy //= s.arbiter.reqs[i]
      s.get[i].ret //= s.mux.in_[i]

    for i in range( num_inports ):
      s.get[i].en //= lambda: s.give.en & ( s.mux.sel == SelType(i) )

    s.give.rdy //= s.granted_get_rdy

  def line_trace( s ):
    in_trace  = '|'.join( [ str(p) for p in s.get ] )
    hold      = ''.join([ '^' if h else '.' for h in s.hold ])
    out_trace = f'{s.give}'
    return f'{in_trace}({hold}){out_trace}'




'''
==========================================================================
MflitPacket_test.py
==========================================================================
Unit tests for MflitPacket.

Author : Yanghui Ou
  Date : Jan 31, 2019
'''
from pymtl3 import *

from .MflitPacket import MflitPacket

@bitstruct
class SimpleFormat:
  opaque : Bits8
  src    : Bits8
  dst    : Bits8
  plen   : Bits8

def test_add():
  # An empty packet
  pkt = MflitPacket( SimpleFormat )
  assert pkt.empty()

  header      = b32(0) 
  header[0:8] = b8(3)
  pkt.add( header )
  pkt.add( b32(0xdeadbeef) )
  assert not pkt.empty()
  pkt.add( b32(0xfaceb00c) )
  assert not pkt.empty()
  pkt.add( b32(0xfeedbabe) )
  assert not pkt.empty()

  assert pkt.full()
  assert pkt.flits[1] == 0xdeadbeef
  assert pkt.flits[2] == 0xfaceb00c
  assert pkt.flits[3] == 0xfeedbabe

def test_pop():
  # A single-flit packet
  header = b32(0)
  pkt    = MflitPacket( SimpleFormat, [ header ] )
  assert not pkt.empty()
  assert pkt.full()
  assert pkt.pop() == b32(0)
  assert pkt.empty()
  assert not pkt.full()

  # A multi-flit packet
  header      = b32(0)
  header[0:8] = b8(2)
  flits       = [ header, b32(0xfaceb00c), b32(0x8badf00d) ]
  pkt         = MflitPacket( SimpleFormat, flits )

  assert not pkt.empty()
  assert pkt.full()

  assert pkt.pop() == b32(2)
  assert not pkt.full()
  assert not pkt.empty()

  assert pkt.pop() == b32(0xfaceb00c)
  assert not pkt.full()
  assert not pkt.empty()

  assert pkt.pop() == b32(0x8badf00d)
  assert not pkt.full()
  assert pkt.empty()



"""
==========================================================================
CMeshNetworkRTL.py
==========================================================================
RTL implementation of concetrated mesh network.

Author : Cheng Tan
  Date : Mar 10, 2019
"""
from channel.ChannelRTL import ChannelRTL
from pymtl3 import *
from pymtl3.stdlib.ifcs.SendRecvIfc import *

from .CMeshRouterRTL import CMeshRouterRTL
from .directions import *


class CMeshNetworkRTL( Component ):
  def construct( s, PacketType, PositionType,
                 ncols=4, nrows=4, num_nodes_each=4, chl_lat=0 ):

    # Constants

    s.num_routers   = ncols * nrows
    s.num_terminals = s.num_routers * num_nodes_each
    num_channels    = (nrows*(ncols-1)+ncols*(nrows-1)) * 2
    num_inports     = 4 + num_nodes_each
    num_outports    = 4 + num_nodes_each
    XType           = mk_bits( clog2(ncols) )
    YType           = mk_bits( clog2(nrows ) )

    # Interface

    s.recv = [ RecvIfcRTL(PacketType) for _ in range( s.num_terminals ) ]
    s.send = [ SendIfcRTL(PacketType) for _ in range( s.num_terminals ) ]

    # Components

    s.routers  = [ CMeshRouterRTL( PacketType, PositionType, num_inports,
                     num_outports ) for i in range( s.num_routers ) ]

    s.channels = [ ChannelRTL( PacketType, latency = chl_lat)
                   for _ in range( num_channels ) ]

    # Connect s.routers together in Mesh
    # FIXME: we need to calculate the bit width for directions. Currently
    # the translation pass may throw an error.

    chl_id  = 0
    for i in range( s.num_routers ):
      if i // ncols > 0:
        s.routers[i].send[SOUTH] //= s.channels[chl_id].recv
        s.channels[chl_id].send  //= s.routers[i-ncols].recv[NORTH]
        chl_id += 1

      if i // ncols < nrows - 1:
        s.routers[i].send[NORTH] //= s.channels[chl_id].recv
        s.channels[chl_id].send  //= s.routers[i+ncols].recv[SOUTH]
        chl_id += 1

      if i % ncols > 0:
        s.routers[i].send[WEST] //= s.channels[chl_id].recv
        s.channels[chl_id].send //= s.routers[i-1].recv[EAST]
        chl_id += 1

      if i % ncols < ncols - 1:
        s.routers[i].send[EAST] //= s.channels[chl_id].recv
        s.channels[chl_id].send //= s.routers[i+1].recv[WEST]
        chl_id += 1

      # Connect the self port (with Network Interface)

      for j in range( num_nodes_each ):
        ifc_idx = i * num_nodes_each + j
        s.recv[ ifc_idx ] //= s.routers[ i ].recv[ 4 + j ]
        s.send[ ifc_idx ] //= s.routers[ i ].send[ 4 + j ]

      # Connect the unused ports

      if i // ncols == 0:
        s.routers[i].send[SOUTH].rdy         //= 0
        s.routers[i].recv[SOUTH].en          //= 0
        s.routers[i].recv[SOUTH].msg.payload //= 0

      if i // ncols == nrows - 1:
        s.routers[i].send[NORTH].rdy         //= 0
        s.routers[i].recv[NORTH].en          //= 0
        s.routers[i].recv[NORTH].msg.payload //= 0

      if i % ncols == 0:
        s.routers[i].send[WEST].rdy          //=  0
        s.routers[i].recv[WEST].en           //=  0
        s.routers[i].recv[WEST].msg.payload  //=  0

      if i % ncols == ncols - 1:
        s.routers[i].send[EAST].rdy          //=  0
        s.routers[i].recv[EAST].en           //=  0
        s.routers[i].recv[EAST].msg.payload  //=  0

#    # FIXME: unable to connect a struct to a port.
#    @s.update
#    def up_pos():
    for y in range( nrows ):
      for x in range( ncols ):
#        s.routers[y*ncols+x].pos = PositionType( x, y )
        s.routers[y*ncols+x].pos.pos_x //= XType(x)
        s.routers[y*ncols+x].pos.pos_y //= YType(y)

  def line_trace( s ):
    trace = [ f"{s.send[i]}" for i in range( s.num_terminals ) ]
    return "|".join( trace )



"""
=========================================================================
RingNetworkRTL.py
=========================================================================
Ring network implementation.

Author : Yanghui Ou, Cheng Tan
  Date : June 22, 2019
"""
from ocnlib.ifcs.CreditIfc import (CreditRecvRTL2SendRTL,
                                      RecvRTL2CreditSendRTL)
from pymtl3 import *
from pymtl3.stdlib.ifcs import RecvIfcRTL, SendIfcRTL

from .directions import *
from .RingRouterRTL import RingRouterRTL


class RingNetworkRTL( Component ):

  def construct( s,
    PacketType,
    PositionType,
    num_routers=4,
    chl_lat=0,
    vc=2,
    credit_line=2,
   ):

    # Constants
    s.num_routers = num_routers
#    IDType = mk_bits(clogs(num_routers))

    # Interface
    s.recv = [ RecvIfcRTL(PacketType) for _ in range(s.num_routers)]
    s.send = [ SendIfcRTL(PacketType) for _ in range(s.num_routers)]

    # Components
    s.routers = [ RingRouterRTL( PacketType, PositionType, num_routers, vc=vc )
                     for i in range( num_routers ) ]

    s.recv_adp = [ RecvRTL2CreditSendRTL( PacketType, vc=vc, credit_line=credit_line )
                        for _ in range( num_routers ) ]
    s.send_adp = [ CreditRecvRTL2SendRTL( PacketType, vc=vc, credit_line=credit_line )
                        for _ in range( num_routers ) ]

    # Connect s.routers together in ring

    for i in range( s.num_routers ):
      next_id = (i+1) % num_routers
      s.routers[i].send[RIGHT]      //= s.routers[next_id].recv[LEFT]
      s.routers[next_id].send[LEFT] //= s.routers[i].recv[RIGHT]
      # Connect the self port (with Network Interface)
      s.recv[i]                     //= s.recv_adp[i].recv
      s.recv_adp[i].send            //= s.routers[i].recv[SELF]
      s.routers[i].send[SELF]       //= s.send_adp[i].recv
      s.send_adp[i].send            //= s.send[i]

    @s.update
    def up_pos():
      for r in range( s.num_routers ):
        s.routers[r].pos = PositionType( r )

  def line_trace( s, level='1pkt' ):

    # Verbose line trace
    if level == 'verbose':
      #  Input unit and route unit
      in_trace = [ "" for _ in range( s.num_routers ) ]
      for i in range ( s.num_routers ):
        in_trace[i] += str(s.recv[i])
        for j in range( 3 ):
          in_trace[i] += "IU{}{{".format( j )
          in_trace[i] += "({},{})".format( s.routers[i].input_units[j].buffers[0].count, s.routers[i].input_units[j].buffers[1].count )
          in_trace[i] += ">{}}}".format( "|".join( [ str(s.routers[i].input_units[j].give[x].msg) for x in range(2)] ) )
          in_trace[i] += "RU{}{{".format( j )
          in_trace[i] += "|".join([ "{}r{}".format( s.routers[i].route_units[j].give[x].msg, s.routers[i].route_units[j].give[x].rdy ) for x in range(3) ])
          in_trace[i] += "}}"

      # Switch unit and output unit
      out_trace = [ "" for _ in range( s.num_routers ) ]
      for i in range( s.num_routers ):
        for j in range( 3 ):
          out_trace[i] += "SU{}{{".format( j )
          out_trace[i] += "|".join([ "{}r{}".format( s.routers[i].switch_units[j].get[x].msg, s.routers[i].switch_units[j].get[x].rdy ) ])
          out_trace[i] += "}}"
          out_trace[i] += "OU{}{{".format( j )
          out_trace[i] += "({}r{})".format( s.routers[i].output_units[j].get.ret, s.routers[i].output_units[j].get.rdy )
          out_trace[i] += "<{},{}>".format( s.routers[i].output_units[j].credit[0].count, s.routers[i].output_units[j].credit[1].count )
        out_trace[i] += str(s.send[i])
      # Assemble line trace
      in_str  = " | ".join( in_trace )
      out_str = " | ".join( out_trace )
      return  " >>>>> ".join( [ "{} __XX__ {}".format( in_trace[i], out_trace[i] ) for i in range(s.num_routers)] )

    # Simple line trace
    elif level=='simple':
      in_trace  = [ str(s.recv[i]) for i in range( s.num_routers) ]
      out_trace = [ str(s.send[i]) for i in range( s.num_routers) ]
      in_str = "|".join( in_trace )
      out_str = "|".join( out_trace )
      return "{}(){}".format( in_str, out_str )

    else:
      send_lst = []
      recv_lst = []
      for r in s.routers:
        has_recv = any([ r.recv[i].en for i in range(3) ])
        has_send = any([ r.send[i].en for i in range(3) ])
        if has_send:
          send_lst.append( f'{r.pos}' )
        if has_recv:
          recv_lst.append( f'{r.pos}')
      send_str = ','.join( send_lst )
      recv_str = ','.join( recv_lst )
      return f' {send_str:2} -> {recv_str:2}'



"""
=========================================================================
MeshRouteRTL.py
=========================================================================
Simple network-on-chip router, try to connect all the units together

Author : Cheng Tan, Yanghui Ou
  Date : Mar 8, 2019
"""
from pymtl3 import *
from router.InputUnitRTL import InputUnitRTL
from router.OutputUnitRTL import OutputUnitRTL
from router.Router import Router
from router.SwitchUnitRTL import SwitchUnitRTL

from .DORYMeshRouteUnitRTL import DORYMeshRouteUnitRTL


class MeshRouterRTL( Router ):

  def construct( s, PacketType, PositionType, InputUnitType = InputUnitRTL,
                 RouteUnitType = DORYMeshRouteUnitRTL,
                 SwitchUnitType = SwitchUnitRTL ):

    super().construct(
      PacketType, PositionType, 5, 5, InputUnitType, RouteUnitType,
      SwitchUnitType, OutputUnitRTL )



"""
=========================================================================
SwitchUnitRTL_test.py
=========================================================================
Test for SwitchUnitRTL.

 Author : Yanghui Ou, Cheng Tan
   Date : June 22, 2019
"""
from ocnlib.ifcs.packets import mk_generic_pkt
from pymtl3 import *
from pymtl3.stdlib.test.test_sinks import TestSinkCL
from pymtl3.stdlib.test.test_srcs import TestSrcCL
from router.SwitchUnitRTL import SwitchUnitRTL

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

def test_switch_unit_simpe():
  dut = SwitchUnitRTL( Bits32, num_inports=5 )
  dut.elaborate()
  dut.apply( SimulationPass() )
  dut.sim_reset()

  print("")
  dut.get[0].rdy = b1(1)
  dut.get[0].ret = b32(0xfaceb00c)
  dut.get[1].rdy = b1(1)
  dut.get[1].ret = b32(0xdeadface)
  dut.get[4].rdy = b1(1)
  dut.get[4].ret = b32(0xbaadbeef)
  dut.give.en = b1(0)
  dut.eval_combinational()
  dut.tick()
  print( dut.line_trace() )

  for i in range( 3 ):
    assert dut.give.rdy
    assert dut.give.ret in { b32(0xfaceb00c), b32(0xdeadface), b32(0xbaadbeef)  }
    dut.give.en = b1(1)
    dut.eval_combinational()
    dut.tick()
    print( dut.line_trace() )



"""
==========================================================================
DORYTorusRouteUnitRTL_test.py
==========================================================================
Test for DORYTorusRouteUnitRTL

Author : Yanghui Ou
  Date : June 28, 2019
"""
from itertools import product

import pytest

from ocnlib.ifcs.packets import mk_mesh_pkt
from ocnlib.ifcs.positions import mk_mesh_pos
from ocnlib.utils import run_sim
from ocnlib.test.net_sinks import TestNetSinkRTL
from pymtl3 import *
from pymtl3.stdlib.rtl.queues import BypassQueueRTL
from pymtl3.stdlib.test.test_srcs import TestSrcRTL
from torusnet.DORYTorusRouteUnitRTL import DORYTorusRouteUnitRTL
from torusnet.RouteUnitDorFL import RouteUnitDorFL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, PktType, src_msgs, sink_msgs, ncols=2, nrows=2, pos_x=0, pos_y=0 ):

    outports = 5
    MeshPos = mk_mesh_pos( ncols, nrows )

    match_func = lambda a, b : a.src_x == b.src_x and a.src_y == b.src_y and \
                               a.dst_y == b.dst_y and a.payload == b.payload


    s.src   = TestSrcRTL( PktType, src_msgs )
    s.src_q = BypassQueueRTL( PktType, num_entries=1 )
    s.dut   = DORYTorusRouteUnitRTL( PktType, MeshPos, ncols=ncols, nrows=nrows )
    s.sinks = [ TestNetSinkRTL( PktType, sink_msgs[i], match_func=match_func )
                for i in range ( outports ) ]

    # Connections
    s.src.send  //= s.src_q.enq
    s.src_q.deq //= s.dut.get

    for i in range ( s.dut.num_outports ):
      s.dut.give[i].ret //= s.sinks[i].recv.msg

    @s.update
    def up_give_en():
      for i in range (s.dut.num_outports):
        if s.dut.give[i].rdy and s.sinks[i].recv.rdy:
          s.dut.give[i].en   = b1(1)
          s.sinks[i].recv.en = b1(1)
        else:
          s.dut.give[i].en   = b1(0)
          s.sinks[i].recv.en = b1(0)

    @s.update
    def up_dut_pos():
      s.dut.pos = MeshPos( pos_x, pos_y )

  def done( s ):
    sinks_done = 1
    for i in range( s.dut.num_outports ):
      if not s.sinks[i].done():
        sinks_done = 0
    return s.src.done() and sinks_done

  def line_trace( s ):
    return "{}".format( s.dut.line_trace() )

#-------------------------------------------------------------------------
# mk_dst_pkts
#-------------------------------------------------------------------------
# A helper function that computes destination packets using the FL model.

def mk_dst_pkts( pos_x, pos_y, ncols, nrows, src_pkts ):
  route_unit = RouteUnitDorFL( pos_x, pos_y, ncols, nrows, dimension='y' )
  return route_unit.route( src_pkts )

#=========================================================================
# Test cases
#=========================================================================
# TODO: Test DORX as well.

class RouteUnitDorRTL_Tests:

  @classmethod
  def setup_method( cls ):
    pass

  @pytest.mark.parametrize(
    'pos_x, pos_y',
    product( [ 0, 1, 2, 3 ], [ 0, 1, 2, 3 ] )
  )
  def test_simple_4x4( s, pos_x, pos_y ):

    ncols = 4
    nrows  = 4

    Pkt = mk_mesh_pkt( ncols, nrows, vc=2 )

    src_pkts = [
      #   src_x  y  dst_x  y  opq  vc  payload
      Pkt(    0, 0,     1, 1,   0,  0, 0xfaceb00c ),
      Pkt(    0, 0,     0, 0,   0,  0, 0xdeaddead ),
      Pkt(    0, 0,     1, 0,   0,  0, 0xdeadface ),
      Pkt(    0, 0,     3, 3,   0,  0, 0xdeadface ),
    ]
    dst_pkts = mk_dst_pkts( pos_x, pos_y, ncols, nrows, src_pkts )
    th = TestHarness( Pkt, src_pkts, dst_pkts, ncols, nrows, pos_x, pos_y )
    run_sim( th )



'''
==========================================================================
XbarMflitRTL_test.py
==========================================================================
Test cases for the multi-flit xbar.

Author : Yanghui Ou
  Date : Feb 19, 2020
'''
import pytest
import hypothesis
from hypothesis import strategies as st
from pymtl3 import *
from pymtl3.stdlib.test import mk_test_case_table
from ocnlib.utils import to_bitstruct, run_sim
from ocnlib.packets import MflitPacket as Packet
from ocnlib.test.test_srcs import MflitPacketSourceRTL as TestSource
from ocnlib.test.test_sinks import MflitPacketSinkRTL as TestSink
from ..XbarMflitRTL import XbarMflitRTL

#-------------------------------------------------------------------------
# TestHeader
#-------------------------------------------------------------------------

@bitstruct
class TestHeader:
  src    : Bits4
  dst    : Bits4
  opaque : Bits4
  plen   : Bits4

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, Header, num_inports, num_outports, pkts ):

    src_pkts, sink_pkts = \
      arrange_src_sink_pkts( Header, num_inports, num_outports, pkts )

    s.src  = [ TestSource( Header, src_pkts[i] ) for i in range( num_inports ) ]
    s.dut  = XbarMflitRTL( Header, num_inports, num_outports )
    s.sink = [ TestSink( Header, sink_pkts[i] ) for i in range( num_outports ) ]

    for i in range( num_inports ):
      s.src[i].send //= s.dut.recv[i]

    for i in range( num_outports ):
      s.dut.send[i] //= s.sink[i].recv

  def done( s ):
    src_done = True
    sink_done = True
    for m in s.src:
      src_done &= m.done()
    for m in s.sink:
      sink_done &= m.done()
    return src_done and sink_done

  def line_trace( s ):
    return s.dut.line_trace()

#-------------------------------------------------------------------------
# mk_pkt
#-------------------------------------------------------------------------

def mk_pkt( src, dst, payload=[], opaque=0 ):
  plen        = len( payload )
  header      = TestHeader( src, dst, opaque, plen )
  header_bits = to_bits( header )
  flits       = [ header_bits ] + payload
  return Packet( TestHeader, flits )

#-------------------------------------------------------------------------
# arrange_src_sink_pkts
#-------------------------------------------------------------------------

def arrange_src_sink_pkts( Header, num_inports, num_outports, pkt_lst ):
  src_pkts  = [ [] for _ in range( num_inports  ) ]
  sink_pkts = [ [] for _ in range( num_outports ) ]

  for pkt in pkt_lst:
    header = to_bitstruct( pkt.flits[0], Header )
    src    = header.src.uint()
    dst    = header.dst.uint()
    src_pkts [ src ].append( pkt )
    sink_pkts[ dst ].append( pkt )

  return src_pkts, sink_pkts

#-------------------------------------------------------------------------
# test case: basic
#-------------------------------------------------------------------------

def basic_pkts( num_inports, num_outports ):
  return [
    #       src            dst             payload
    mk_pkt( 0,             0,              [                ] ),
    mk_pkt( 0,             num_outports-1, [                ] ),
    mk_pkt( num_inports-1, 0,              [ 0xface, 0xf00d ] ),
  ]

#-------------------------------------------------------------------------
# test case: neighbor
#-------------------------------------------------------------------------

def neighbor_pkts( num_inports, num_outports ):
  pkts = []
  for i in range( num_inports ):
    src = i
    dst = ( i+1 ) % num_outports
    plen = i % 10
    pkts.append( mk_pkt( src, dst, [ x+0xbad0 for x in range( plen ) ] ) )

  for i in range( num_inports ):
    src = i
    dst = ( i+1 ) % num_outports
    plen = ( i + 8 ) % 13
    pkts.append( mk_pkt( src, dst, [ x+0xace0 for x in range( plen ) ] ) )

  return pkts

#-------------------------------------------------------------------------
# test case: product
#-------------------------------------------------------------------------

def product_pkts( num_inports, num_outports ):
  pkts = []
  for i in range( num_inports ):
    for j in range( num_outports ):
      src = i
      dst = j
      plen = ( i + j ) % 16
      pkts.append( mk_pkt( src, dst, [ x+0xfee for x in range( plen ) ] ) )
      pkts.append( mk_pkt( src, dst, [ x+0xdad for x in range( plen ) ] ) )
  return pkts

#-------------------------------------------------------------------------
# test case table
#-------------------------------------------------------------------------

test_cases = [
  (                      'msg_func         n_in n_out init f_intv p_intv' ),
  [ 'basic2x2',           basic_pkts,      2,   2,    0,   0,     0       ],
  [ 'basic2x1',           basic_pkts,      2,   1,    0,   0,     0       ],
  [ 'basic2x9',           basic_pkts,      2,   9,    0,   0,     0       ],
  [ 'basic1x2',           basic_pkts,      1,   2,    0,   0,     0       ],
  [ 'neighbor4x4',        neighbor_pkts,   4,   4,    0,   0,     0       ],
  [ 'neighbor5x2',        neighbor_pkts,   5,   2,    0,   0,     0       ],
  [ 'neighbor3x3_delay',  neighbor_pkts,   4,   4,    5,   2,     9       ],
  [ 'product4x4',         product_pkts,    4,   4,    0,   0,     4       ],
  [ 'product8x8',         product_pkts,    8,   8,    0,   0,     2       ],
  [ 'product3x4',         product_pkts,    3,   4,    0,   0,     1       ],
  [ 'product7x3',         product_pkts,    7,   3,    0,   0,     1       ],
]

test_case_table = mk_test_case_table( test_cases )

#-------------------------------------------------------------------------
# test driver
#-------------------------------------------------------------------------

@pytest.mark.parametrize( **test_case_table )
def test_mflit_xbar( test_params, test_verilog ):
  pkts = test_params.msg_func( test_params.n_in, test_params.n_out )
  trans_backend = 'verilog' if test_verilog else ''
  th = TestHarness( TestHeader, test_params.n_in, test_params.n_out, pkts )
  th.set_param( 'top.sink*.construct',
    initial_delay         = test_params.init,
    flit_interval_delay   = test_params.f_intv,
    packet_interval_delay = test_params.p_intv,
  )
  run_sim( th, translation=trans_backend )

#-------------------------------------------------------------------------
# packet strategy
#-------------------------------------------------------------------------

hex_words = [ 0xf00d, 0xfade, 0xface, 0xbeef, 0x8bad, 0xdeaf, 0xbabe ]

@st.composite
def pkt_strat( draw, num_inports, num_outports, max_plen=15 ):
  payload = draw( st.lists( st.sampled_from( hex_words ),
                            min_size=0, max_size=max_plen ) )
  src     = draw( st.integers(0, num_inports-1 ) )
  dst     = draw( st.integers(0, num_outports-1) )
  return mk_pkt( src, dst, payload )

#-------------------------------------------------------------------------
# pyh2 test
#-------------------------------------------------------------------------

@hypothesis.settings( deadline=None, max_examples=50 )
@hypothesis.given(
  num_inports  = st.integers(1, 16),
  num_outports = st.integers(1, 16),
  pkts         = st.data(),
)
def test_pyh2( num_inports, num_outports, pkts, test_verilog ):
  pkts = pkts.draw( st.lists( pkt_strat( num_inports, num_outports ), min_size=1, max_size=100 ) )
  trans_backend = 'verilog' if test_verilog else ''
  th = TestHarness( TestHeader, num_inports, num_outports, pkts )
  run_sim( th, translation=trans_backend )



#=========================================================================
# DORYRouteUnitRTL_test.py
#=========================================================================
# Test for DORYRouteUnitRTL
#
# Author : Yanghui Ou, Cheng Tan
#   Date : Mar 25, 2019

from meshnet.DORYMeshRouteUnitRTL import DORYMeshRouteUnitRTL
from ocnlib.ifcs.packets import mk_mesh_pkt
from ocnlib.ifcs.positions import mk_mesh_pos
from ocnlib.utils import run_sim
from pymtl3 import *
from pymtl3.stdlib.test import TestVectorSimulator
from pymtl3.stdlib.test.test_sinks import TestSinkRTL
from pymtl3.stdlib.test.test_srcs import TestSrcRTL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, src_msgs, sink_msgs, src_initial,
                 src_interval, sink_initial, sink_interval,
                 arrival_time=None ):

    ncols = 4
    nrows = 4

    MeshPos = mk_mesh_pos( ncols, nrows )
    s.dut = DORYMeshRouteUnitRTL( MsgType, MeshPos )
    s.dut.pos = MeshPos( 1, 1 )

    s.src   = TestSrcRTL   ( MsgType, src_msgs,  src_initial,  src_interval  )
    s.sinks = [ TestSinkRTL  ( MsgType, sink_msgs[i], sink_initial, sink_interval, arrival_time[i] )
                for i in range ( s.dut.num_outports ) ]

    # Connections
    s.src.send.msg //= s.dut.get.ret

    for i in range ( s.dut.num_outports ):
      s.dut.give[i].ret //= s.sinks[i].recv.msg

    @s.update
    def up_give_en():
      for i in range (s.dut.num_outports):
        if s.dut.give[i].rdy and s.sinks[i].recv.rdy:
          s.dut.give[i].en   = 1
          s.sinks[i].recv.en = 1
        else:
          s.dut.give[i].en   = 0
          s.sinks[i].recv.en = 0

    # FIXME: connect send to get
    # s.connect( s.src.send.rdy, Bits1( 1 )    )
    # s.connect( s.dut.get.rdy,  s.src.send.en )

  def done( s ):
    sinks_done = 1
    for i in range( s.dut.num_outports ):
      if s.sinks[i].done() == 0:
        sinks_done = 0
    return s.src.done() and sinks_done

  def line_trace( s ):
    return s.src.line_trace() + "-> | " + s.dut.line_trace() + \
                               " | -> " + s.sinks[0].line_trace()

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

#               x,y,pl,dir
test_msgs   = [(0,0,101,0), (0,2,102,1), (0,1,103,2), (2,1,104,3),
               (1,1,105,4), (1,1,106,4)]
result_msgs = [ [], [], [], [], [] ]

arrival_time = [ [1], [2], [3], [4], [5,6] ]

# def test_normal_simple():
#
#   src_packets = []
#   for ( dst_x, dst_y, payload, dir_out ) in test_msgs:
#     pkt = mk_pkt (0, 0, dst_x, dst_y, 1, payload)
#     src_packets.append( pkt )
#     result_msgs[dir_out].append ( pkt )
#
#   th = TestHarness( Packet, src_packets, result_msgs, 0, 0, 0, 0,
#                     arrival_time )
#   run_sim( th )
